#!/usr/bin/env python
# udp_session.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (udp_session.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
.. module:: udp_session.

BitDust udp_session() Automat


EVENTS:
    * :red:`datagram-received`
    * :red:`init`
    * :red:`send-keep-alive`
    * :red:`shutdown`
    * :red:`timer-10sec`
    * :red:`timer-1sec`
    * :red:`timer-20sec`
    * :red:`timer-30sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

import os
import sys
import time

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng
from bitdust.lib import misc
from bitdust.lib import udp

from bitdust.automats import automat

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 14

#------------------------------------------------------------------------------

MIN_PROCESS_SESSIONS_DELAY = 0.001
MAX_PROCESS_SESSIONS_DELAY = 1.0

#------------------------------------------------------------------------------

_SessionsDict = {}
_SessionsDictByPeerAddress = {}
_SessionsDictByPeerID = {}
_KnownPeersDict = {}
_KnownUserIDsDict = {}
_PendingOutboxFiles = []
_ProcessSessionsTask = None
_ProcessSessionsDelay = MIN_PROCESS_SESSIONS_DELAY

#------------------------------------------------------------------------------


def sessions():
    global _SessionsDict
    return _SessionsDict


def sessions_by_peer_address():
    global _SessionsDictByPeerAddress
    return _SessionsDictByPeerAddress


def sessions_by_peer_id():
    global _SessionsDictByPeerID
    return _SessionsDictByPeerID


def pending_outbox_files():
    global _PendingOutboxFiles
    return _PendingOutboxFiles


#------------------------------------------------------------------------------


def create(node, peer_address, peer_id=None):
    if _Debug:
        lg.out(_DebugLevel, 'udp_session.create peer_address=%r' % peer_address)
    s = UDPSession(node, peer_address, peer_id)
    sessions()[s.id] = s
    try:
        sessions_by_peer_address()[peer_address].append(s)
    except:
        sessions_by_peer_address()[peer_address] = [
            s,
        ]
    if peer_id:
        try:
            sessions_by_peer_id()[peer_id].append(s)
        except:
            sessions_by_peer_id()[peer_id] = [
                s,
            ]
    return s


def get(peer_address):
    return sessions_by_peer_address().get(peer_address, [])


def get_by_peer_id(peer_id):
    return sessions_by_peer_id().get(peer_id, [])


def close(peer_address):
    active_sessions = get(peer_address)
    if not active_sessions:
        return False
    for s in active_sessions:
        s.automat('shutdown')
    return True


def add_pending_outbox_file(filename, host, description='', result_defer=None, keep_alive=True):
    pending_outbox_files().append((filename, host, description, result_defer, keep_alive, time.time()))
    if _Debug:
        lg.out(_DebugLevel, 'udp_session.add_pending_outbox_file %s for %s : %s' % (os.path.basename(filename), host, description))


def remove_pending_outbox_file(host, filename):
    ok = False
    i = 0
    while i < len(pending_outbox_files()):
        fn, hst, description, result_defer, keep_alive, tm = pending_outbox_files()[i]
        if fn == filename and host == hst:
            if _Debug:
                lg.out(_DebugLevel, 'udp_interface.cancel_outbox_file removed pending %s for %s' % (os.path.basename(fn), hst))
            pending_outbox_files().pop(i)
            ok = True
        else:
            i += 1
    return ok


def report_and_remove_pending_outbox_files_to_host(remote_host, error_message):
    from bitdust.transport.udp import udp_interface
    global _PendingOutboxFiles
    i = 0
    while i < len(_PendingOutboxFiles):
        filename, host, description, result_defer, keep_alive, tm = _PendingOutboxFiles[i]
        if host != remote_host:
            i += 1
            continue
        try:
            udp_interface.interface_cancelled_file_sending(remote_host, filename, 0, description, error_message).addErrback(lambda err: lg.exc(err))
        except Exception as exc:
            lg.warn(str(exc))
        if result_defer:
            result_defer.callback(((filename, description), 'failed', error_message))
        _PendingOutboxFiles.pop(i)


def process_sessions(sessions_to_process=None):
    global _ProcessSessionsTask
    global _ProcessSessionsDelay
    has_activity = False
    if not sessions_to_process:
        sessions_to_process = list(sessions().values())
    for s in sessions_to_process:
        if not s.peer_id:
            continue
        if not s.file_queue:
            continue
        if s.state != 'CONNECTED':
            continue
        has_outbox = s.file_queue.process_outbox_queue()
        has_sends = s.file_queue.process_outbox_files()
        if has_sends or has_outbox:
            has_activity = True
    if _ProcessSessionsTask is None or _ProcessSessionsTask.called:
        if has_activity:
            _ProcessSessionsDelay = MIN_PROCESS_SESSIONS_DELAY
        else:
            _ProcessSessionsDelay = misc.LoopAttenuation(
                _ProcessSessionsDelay,
                has_activity,
                MIN_PROCESS_SESSIONS_DELAY,
                MAX_PROCESS_SESSIONS_DELAY,
            )
        # attenuation
        _ProcessSessionsTask = reactor.callLater(  # @UndefinedVariable
            _ProcessSessionsDelay,
            process_sessions,
        )


def stop_process_sessions():
    global _ProcessSessionsTask
    if _ProcessSessionsTask:
        if _ProcessSessionsTask.active():
            _ProcessSessionsTask.cancel()
        _ProcessSessionsTask = None


#------------------------------------------------------------------------------


class UDPSession(automat.Automat):
    """
    This class implements all the functionality of the ``udp_session()`` state
    machine.
    """

    fast = True

    timers = {
        'timer-1sec': (1.0, ['PING', 'GREETING']),
        'timer-30sec': (30.0, ['GREETING']),
        'timer-10sec': (10.0, ['CONNECTED']),
        'timer-20sec': (20.0, ['PING']),
    }

    MESSAGES = {
        'MSG_1': 'remote peer is not active',
        'MSG_2': 'greeting is timed out',
        'MSG_3': 'ping remote machine has failed',
        'MSG_4': 'session has been closed at startup',
    }

    def __init__(self, node, peer_address, peer_id=None):
        from bitdust.transport.udp import udp_file_queue
        self.node = node
        self.peer_address = peer_address
        self.peer_id = peer_id
        self.peer_idurl = None
        self.bytes_sent = 0
        self.bytes_received = 0
        self.file_queue = udp_file_queue.FileQueue(self)
        name = 'udp_session[%s:%d:%s]' % (self.peer_address[0], self.peer_address[1], str(self.peer_id))
        automat.Automat.__init__(self, name, 'AT_STARTUP', debug_level=_DebugLevel, log_events=_Debug)

    def is_connected(self):
        return self.state in [
            'CONNECTED',
        ]

    def get_proto(self):
        return 'udp'

    def get_host(self):
        return self.peer_id

    def get_idurl(self):
        return self.peer_idurl

    def msg(self, msgid, *args, **kwargs):
        return self.MESSAGES.get(msgid, '')

    def init(self):
        """
        Method to initialize additional variables and flags at creation of the
        state machine.
        """
        if _Debug:
            self.log_events = True
            self.log_transitions = True
        self.last_datagram_received_time = 0
        self.my_rtt_id = '0'  # out
        self.peer_rtt_id = '0'  # in
        self.rtts = {}
        self.min_rtt = None

    def send_packet(self, command, payload):
        self.bytes_sent += len(payload)
        return udp.send_command(self.node.listen_port, command, payload, self.peer_address)

    def A(self, event, *args, **kwargs):
        #---AT_STARTUP---
        if self.state == 'AT_STARTUP':
            if event == 'init':
                self.state = 'PING'
                self.doInit(*args, **kwargs)
                self.doStartRTT(*args, **kwargs)
                self.doPing(*args, **kwargs)
            elif event == 'shutdown':
                self.state = 'CLOSED'
                self.doErrMsg(event, self.msg('MSG_4', *args, **kwargs))
                self.doClosePendingFiles(*args, **kwargs)
                self.doNotifyDisconnected(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---CONNECTED---
        elif self.state == 'CONNECTED':
            if event == 'shutdown' or (event == 'timer-10sec' and not self.isSessionActive(*args, **kwargs)):
                self.state = 'CLOSED'
                self.doErrMsg(event, self.msg('MSG_1', *args, **kwargs))
                self.doClosePendingFiles(*args, **kwargs)
                self.doNotifyDisconnected(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'datagram-received' and self.isGreeting(*args, **kwargs):
                self.doAcceptGreeting(*args, **kwargs)
                self.doFinishRTT(*args, **kwargs)
                self.doAlive(*args, **kwargs)
            elif event == 'datagram-received' and self.isPayloadData(*args, **kwargs):
                self.doReceiveData(*args, **kwargs)
            elif event == 'datagram-received' and self.isPing(*args, **kwargs):
                self.doAcceptPing(*args, **kwargs)
                self.doGreeting(*args, **kwargs)
            elif event == 'send-keep-alive' or event == 'timer-10sec':
                self.doAlive(*args, **kwargs)
        #---PING---
        elif self.state == 'PING':
            if event == 'timer-1sec':
                self.doStartRTT(*args, **kwargs)
                self.doPing(*args, **kwargs)
            elif event == 'datagram-received' and self.isGreeting(*args, **kwargs):
                self.state = 'GREETING'
                self.doAcceptGreeting(*args, **kwargs)
                self.doFinishRTT(*args, **kwargs)
                self.doStartRTT(*args, **kwargs)
                self.doGreeting(*args, **kwargs)
            elif event == 'datagram-received' and self.isPing(*args, **kwargs):
                self.state = 'GREETING'
                self.doAcceptPing(*args, **kwargs)
                self.doStartRTT(*args, **kwargs)
                self.doGreeting(*args, **kwargs)
            elif event == 'shutdown' or event == 'timer-20sec':
                self.state = 'CLOSED'
                self.doErrMsg(event, self.msg('MSG_3', *args, **kwargs))
                self.doClosePendingFiles(*args, **kwargs)
                self.doNotifyDisconnected(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
        #---GREETING---
        elif self.state == 'GREETING':
            if event == 'timer-1sec':
                self.doStartRTT(*args, **kwargs)
                self.doGreeting(*args, **kwargs)
            elif event == 'shutdown' or event == 'timer-30sec':
                self.state = 'CLOSED'
                self.doErrMsg(event, self.msg('MSG_2', *args, **kwargs))
                self.doClosePendingFiles(*args, **kwargs)
                self.doNotifyDisconnected(*args, **kwargs)
                self.doDestroyMe(*args, **kwargs)
            elif event == 'datagram-received' and self.isAlive(*args, **kwargs):
                self.state = 'CONNECTED'
                self.doAcceptAlive(*args, **kwargs)
                self.doFinishAllRTTs(*args, **kwargs)
                self.doNotifyConnected(*args, **kwargs)
                self.doCheckPendingFiles(*args, **kwargs)
                self.doAlive(*args, **kwargs)
            elif event == 'datagram-received' and self.isPing(*args, **kwargs):
                self.doAcceptPing(*args, **kwargs)
                self.doStartRTT(*args, **kwargs)
                self.doGreeting(*args, **kwargs)
            elif event == 'datagram-received' and self.isGreeting(*args, **kwargs):
                self.doAcceptGreeting(*args, **kwargs)
                self.doFinishRTT(*args, **kwargs)
                self.doAlive(*args, **kwargs)
        #---CLOSED---
        elif self.state == 'CLOSED':
            pass
        return None

    def isPayloadData(self, *args, **kwargs):
        """
        Condition method.
        """
        command = args[0][0][0]
        return (command == udp.CMD_DATA or command == udp.CMD_ACK)

    def isPing(self, *args, **kwargs):
        """
        Condition method.
        """
        command = args[0][0][0]
        return (command == udp.CMD_PING)

    def isGreeting(self, *args, **kwargs):
        """
        Condition method.
        """
        command = args[0][0][0]
        return (command == udp.CMD_GREETING)

    def isAlive(self, *args, **kwargs):
        """
        Condition method.
        """
        command = args[0][0][0]
        return (command == udp.CMD_ALIVE)

#    def isGreetingOrAlive(self, *args, **kwargs):
#        """
#        Condition method.
#        """
#        command = arg[0][0]
#        return ( command == udp.CMD_ALIVE or command == udp.CMD_GREETING)

    def isSessionActive(self, *args, **kwargs):
        """
        Condition method.
        """
        return time.time() - self.last_datagram_received_time < 20

    def doInit(self, *args, **kwargs):
        """
        Action method.
        """
        # self.listen_port, self.my_id, self.my_address = arg

    def doPing(self, *args, **kwargs):
        """
        Action method.
        """
        #        if udp_stream._Debug:
        #            if not (self.peer_address.count('37.18.255.42') or self.peer_address.count('37.18.255.38')):
        #                return
        # rtt_id_out = self._rtt_start('PING')
        udp.send_command(
            self.node.listen_port,
            udp.CMD_PING,
            strng.to_bin(self.my_rtt_id),
            self.peer_address,
        )
        # # print 'doPing', self.my_rtt_id
        self.my_rtt_id = '0'

    def doGreeting(self, *args, **kwargs):
        """
        Action method.
        """
        # rtt_id_out = self._rtt_start('GREETING')
        payload = '%s %s %s %s' % (
            str(self.node.my_id),
            str(self.node.my_idurl),
            str(self.peer_rtt_id),
            str(self.my_rtt_id),
        )
        udp.send_command(self.node.listen_port, udp.CMD_GREETING, strng.to_bin(payload), self.peer_address)
        # print 'doGreeting', self.peer_rtt_id, self.my_rtt_id
        self.peer_rtt_id = '0'
        self.my_rtt_id = '0'

    def doAlive(self, *args, **kwargs):
        """
        Action method.
        """
        udp.send_command(self.node.listen_port, udp.CMD_ALIVE, strng.to_bin(self.peer_rtt_id), self.peer_address)
        # print 'doAlive', self.peer_rtt_id
        self.peer_rtt_id = '0'

    def doAcceptPing(self, *args, **kwargs):
        """
        Action method.
        """
        address, command, payload = self._dispatch_datagram(args[0])
        self.peer_rtt_id = payload.strip()
        # print 'doAcceptPing', self.peer_rtt_id

    def doAcceptGreeting(self, *args, **kwargs):
        """
        Action method.
        """
        address, command, payload = self._dispatch_datagram(args[0])
        parts = payload.split(' ')
        try:
            new_peer_id = parts[0]
            new_peer_idurl = parts[1]
            if len(parts) >= 4:
                self.peer_rtt_id = parts[3]
            else:
                self.peer_rtt_id = '0'
            if len(parts) >= 3:
                self.my_rtt_id = parts[2]
            else:
                self.my_rtt_id = '0'
        except:
            lg.exc()
            return
        # print 'doAcceptGreeting', self.peer_rtt_id, self.my_rtt_id
        # self._rtt_finish(rtt_id_in)
        # rtt_id_out = self._rtt_start('ALIVE')
        # udp.send_command(self.node.listen_port, udp.CMD_ALIVE, '', self.peer_address)
        first_greeting = False
        if self.peer_id:
            if new_peer_id != self.peer_id:
                lg.warn('session: %s,  peer_id from GREETING is different: %s' % (self, new_peer_id))
        else:
            lg.out(_DebugLevel, 'udp_session.doAcceptGreeting detected peer id : %s for session %s' % (new_peer_id, self.peer_address))
            self.peer_id = new_peer_id
            self.name = 'udp_session[%s:%d:%s]' % (self.peer_address[0], self.peer_address[1], str(self.peer_id))
            first_greeting = True
            try:
                sessions_by_peer_id()[self.peer_id].append(self)
            except:
                sessions_by_peer_id()[self.peer_id] = [
                    self,
                ]
        if self.peer_idurl:
            if new_peer_idurl != self.peer_idurl:
                lg.warn('session: %s,  peer_idurl from GREETING is different: %s' % (self, new_peer_idurl))
        else:
            if _Debug:
                lg.out(_DebugLevel, 'udp_session.doAcceptGreeting detected peer idurl : %s for session %s' % (new_peer_idurl, self.peer_address))
            self.peer_idurl = new_peer_idurl
            first_greeting = True
        if first_greeting:
            for s in sessions().values():
                if self.id == s.id:
                    continue
                if self.peer_id == s.peer_id:
                    lg.warn('got GREETING from another address, close session %s' % s)
                    s.automat('shutdown')
                    continue
                if self.peer_idurl == s.peer_idurl:
                    lg.warn('got GREETING from another idurl, close session %s' % s)
                    s.automat('shutdown')
                    continue

    def doAcceptAlive(self, *args, **kwargs):
        """
        Action method.
        """
        address, command, payload = self._dispatch_datagram(args[0])
        self.my_rtt_id = payload.strip()
        # print 'doAcceptAlive', self.my_rtt_id

    def doReceiveData(self, *args, **kwargs):
        """
        Action method.
        """
        self.last_datagram_received_time = time.time()
        try:
            datagram, address = args[0]
            command, payload = datagram
        except:
            return
        assert address == self.peer_address
        self.bytes_received += len(payload)
        if command == udp.CMD_DATA:
            self.file_queue.on_received_data_packet(payload)
        elif command == udp.CMD_ACK:
            self.file_queue.on_received_ack_packet(payload)
#        elif command == udp.CMD_PING:
#            pass
#        elif command == udp.CMD_ALIVE:
#            pass
#        elif command == udp.CMD_GREETING:
#            pass

    def doNotifyConnected(self, *args, **kwargs):
        """
        Action method.
        """
        # # print 'CONNECTED!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'

    def doNotifyDisconnected(self, *args, **kwargs):
        """
        Action method.
        """
        # # print 'DISCONNECTED!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!'

    def doCheckPendingFiles(self, *args, **kwargs):
        """
        Action method.
        """
        global _PendingOutboxFiles
        i = 0
        outgoings = 0
        # print 'doCheckPendingFiles', self.peer_id, len(_PendingOutboxFiles)
        while i < len(_PendingOutboxFiles):
            filename, host, description, result_defer, keep_alive, tm = _PendingOutboxFiles[i]
            # print filename, host, description,
            if host == self.peer_id:
                outgoings += 1
                # small trick to speed up service packets - they have a high
                # priority
                if description.startswith('Identity') or description.startswith('Ack'):
                    self.file_queue.insert_outbox_file(filename, description, result_defer, keep_alive)
                else:
                    self.file_queue.append_outbox_file(filename, description, result_defer, keep_alive)
                _PendingOutboxFiles.pop(i)
                # print 'pop'
            else:
                # _PendingOutboxFiles.insert(i, (filename, host, description, result_defer, keep_alive, tm))
                i += 1
                # print 'skip'
        # print len(_PendingOutboxFiles)
        if outgoings > 0:
            reactor.callLater(0, process_sessions)  # @UndefinedVariable

    def doClosePendingFiles(self, *args, **kwargs):
        """
        Action method.
        """
        report_and_remove_pending_outbox_files_to_host(self.peer_id, self.error_message)
        self.file_queue.report_failed_inbox_files(self.error_message)
        self.file_queue.report_failed_outbox_files(self.error_message)
        self.file_queue.report_failed_outbox_queue(self.error_message)

    def doStartRTT(self, *args, **kwargs):
        """
        Action method.
        """
        self.my_rtt_id = self._rtt_start(self.state)

    def doFinishRTT(self, *args, **kwargs):
        """
        Action method.
        """
        self._rtt_finish(self.my_rtt_id)
        self.my_rtt_id = '0'

    def doFinishAllRTTs(self, *args, **kwargs):
        """
        Action method.
        """
        self._rtt_finish(self.my_rtt_id)
        self.my_rtt_id = '0'
        to_remove = []
        good_rtts = {}
        min_rtt = sys.float_info.max
        for rtt_id in self.rtts.keys():
            if self.rtts[rtt_id][1] == -1:
                to_remove.append(rtt_id)
            else:
                rtt = self.rtts[rtt_id][1] - self.rtts[rtt_id][0]
                if rtt < min_rtt:
                    min_rtt = rtt
                good_rtts[rtt_id] = rtt
        self.min_rtt = min_rtt
        for rtt_id in to_remove:
            # print 'doFinishAllRTTs closed', rtt_id
            del self.rtts[rtt_id]
            lg.out(_DebugLevel, 'udp_session.doFinishAllRTTs: %r' % good_rtts)  # print self.rtts.keys()

    def doErrMsg(self, event, *args, **kwargs):
        """
        Action method.
        """
        if event.count('shutdown'):
            self.error_message = 'session has been closed'
        else:
            self.error_message = args[0]

    def doDestroyMe(self, *args, **kwargs):
        """
        Action method.
        """
        if _Debug:
            lg.out(_DebugLevel, 'udp_session.doDestroyMe %s' % self)
        self.file_queue.close()
        self.file_queue = None
        self.node = None
        sessions().pop(self.id)
        sessions_by_peer_address()[self.peer_address].remove(self)
        if len(sessions_by_peer_address()[self.peer_address]) == 0:
            sessions_by_peer_address().pop(self.peer_address)
        if self.peer_id in list(sessions_by_peer_id().keys()):
            sessions_by_peer_id()[self.peer_id].remove(self)
            if len(sessions_by_peer_id()[self.peer_id]) == 0:
                sessions_by_peer_id().pop(self.peer_id)
        self.destroy()

    def _dispatch_datagram(self, *args, **kwargs):
        self.last_datagram_received_time = time.time()
        try:
            datagram, address = args[0]
            command, payload = datagram
        except:
            lg.exc()
            return None, None, None
        assert address == self.peer_address
        return address, command, payload

    def _rtt_start(self, name):
        i = 0
        while name + str(i) in list(self.rtts.keys()):
            i += 1
        new_rtt_id = name + str(i)
        self.rtts[new_rtt_id] = [time.time(), -1]
        #     lg.out(_DebugLevel, 'udp_session._rtt_start added new RTT %s' % new_rtt_id)
        if len(self.rtts) > 10:
            oldest_rtt_moment = time.time()
            oldest_rtt_id = None
            for rtt_id in self.rtts.keys():
                rtt_data = self.rtts[rtt_id]
                if rtt_data[0] < oldest_rtt_moment:
                    oldest_rtt_moment = rtt_data[1]
                    oldest_rtt_id = rtt_id
            if oldest_rtt_id:
                # rtt = self.rtts[oldest_rtt_id][1] - self.rtts[oldest_rtt_id][0]
                del self.rtts[oldest_rtt_id]
                #     lg.out(_DebugLevel, 'udp_session._rtt_start removed oldest RTT %s  %r' % (
                #     oldest_rtt_id, rtt))
        while len(self.rtts) > 10:
            i = self.rtts.popitem()
            #     lg.out(_DebugLevel, 'udp_session._rtt_finish removed one extra item : %r' % str(i))
        # print 'rtt start', new_rtt_id, self.peer_id
        return new_rtt_id

    def _rtt_finish(self, rtt_id_in):
        # print 'rtt finish', rtt_id_in, self.peer_id
        if rtt_id_in == '0' or rtt_id_in not in self.rtts:
            return


#             or rtt_id_in not in self.rtts:
#            for rtt_id in self.rtts.keys():
#                if self.rtts[rtt_id][1] == -1:
#                    rtt = self.rtts[rtt_id][1] - self.rtts[rtt_id][0]
#                    del self.rtts[rtt_id]
#                        lg.out(_DebugLevel, 'udp_session._rtt_finish removed not finished RTT %s:%r' % (
#                        rtt_id, rtt))
#                    return
#            lg.warn('rtt %s not found in %s' % (rtt_id_in, self))
#            return
        self.rtts[rtt_id_in][1] = time.time()
        rtt = self.rtts[rtt_id_in][1] - self.rtts[rtt_id_in][0]
        if _Debug:
            lg.out(_DebugLevel, 'udp_session._rtt_finish registered RTT %s  %r' % (rtt_id_in, rtt))



import os
import re
import random
import hashlib
import threading
import traceback
import time
import json
import math
import sqlite3
import base64
import socks
import socketserver

from Cryptodome.PublicKey import RSA
from Cryptodome.Hash import SHA
from Cryptodome.Signature import PKCS1_v1_5

#------------------------------------------------------------------------------

from twisted.internet.defer import Deferred
from twisted.internet import reactor

#------------------------------------------------------------------------------

from bitdust_forks.Bismuth import connections  # @UnresolvedImport
from bitdust_forks.Bismuth import essentials  # @UnresolvedImport
from bitdust_forks.Bismuth import mining_heavy3  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.main import settings
from bitdust.main import config

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

_DataDirPath = None
_PoolHost = None
_PoolPort = None

#------------------------------------------------------------------------------

# TODO: find solution here, need to create some default address
alt_add = '1aae2cfe5d01acc8d7cbc90fcf8bb715ca24927504d0d8071c0979c7'

mdiff = 10
min_payout = 1
pool_fee = 0
alt_fee = 0
worker_time = 10

new_time = 0
new_diff = 0
new_hash = None

address = None
node_ip = None
node_port = None
# ledger_path_conf = None
key = None
public_key_hashed = None
private_key_readable = None
m_peer_file = None

mempool_db_path = None
shares_db_path = None
archive_db_path = None

#------------------------------------------------------------------------------


def init():
    global _DataDirPath
    global _PoolHost
    global _PoolPort
    _DataDirPath = settings.ServiceDir('bismuth_blockchain')
    _PoolHost = config.conf().getString('services/bismuth-pool/host', '127.0.0.1')
    _PoolPort = config.conf().getInt('services/bismuth-pool/tcp-port', 18525)
    node_host_port = '{}:{}'.format(
        config.conf().getString('services/bismuth-node/host', '127.0.0.1'),
        config.conf().getInt('services/bismuth-node/tcp-port', 15658),
    )
    starting_defer = Deferred()
    node_thread = threading.Thread(target=run, args=(starting_defer, _DataDirPath, node_host_port, _Debug))
    node_thread.start()
    if _Debug:
        lg.args(_DebugLevel, data_dir_path=_DataDirPath)
    return starting_defer


def shutdown():
    if _Debug:
        lg.dbg(_DebugLevel, '')
    return True


#------------------------------------------------------------------------------


def run(starting_defer, data_dir_path, node_address, verbose=False):
    global _DataDirPath
    global _PoolHost
    global _PoolPort
    # global ledger_path_conf
    global node_ip
    global node_port
    global address
    global key
    global public_key_hashed
    global private_key_readable
    global m_peer_file
    global mempool_db_path
    global shares_db_path
    global archive_db_path

    _DataDirPath = data_dir_path
    if not os.path.exists(data_dir_path):
        os.makedirs(data_dir_path)

    pool_key_path = os.path.join(data_dir_path, 'wallet_key.json')

    m_peer_file = os.path.join(data_dir_path, 'peers.json')

    mempool_db_path = os.path.join(data_dir_path, 'mempool.db')
    shares_db_path = os.path.join(data_dir_path, 'shares.db')
    archive_db_path = os.path.join(data_dir_path, 'archive.db')

    node_ip = node_address.split(':')[0]
    node_port = node_address.split(':')[1]

    if os.path.isfile(pool_key_path):
        if _Debug:
            lg.dbg(_DebugLevel, 'found %s key file' % pool_key_path)

    else:
        # generate key pair and an address
        key = RSA.generate(4096)
        private_key_readable = str(key.exportKey().decode('utf-8'))
        public_key_readable = str(key.publickey().exportKey().decode('utf-8'))
        address = hashlib.sha224(public_key_readable.encode('utf-8')).hexdigest()  # hashed public key
        essentials.keys_save(private_key_readable, public_key_readable, address, pool_key_path)
        if _Debug:
            lg.dbg(_DebugLevel, 'generated new %s key file' % pool_key_path)

    key, public_key_readable, private_key_readable, encrypted, unlocked, public_key_hashed, address, keyfile = essentials.keys_load(wallet_filename=pool_key_path)

    if _Debug:
        lg.args(_DebugLevel, pool_wallet_address=address)

    if not os.path.exists(shares_db_path):
        shares = sqlite3.connect(shares_db_path)
        shares.text_factory = str
        s = shares.cursor()
        execute(s, 'CREATE TABLE IF NOT EXISTS shares (address, shares, timestamp, paid, rate, name, workers, subname)')
        execute(s, 'CREATE TABLE IF NOT EXISTS nonces (nonce)')
        s.close()
        if _Debug:
            lg.dbg(_DebugLevel, 'created shares DB in %r' % shares_db_path)

    if not os.path.exists(archive_db_path):
        archive = sqlite3.connect(archive_db_path)
        archive.text_factory = str
        a = archive.cursor()
        execute(a, 'CREATE TABLE IF NOT EXISTS shares (address, shares, timestamp, paid, rate, name, workers, subname)')
        a.close()
        if _Debug:
            lg.dbg(_DebugLevel, 'created archive DB in %r' % archive_db_path)

    try:
        # background_thread = threading.Thread(target=paydb)
        # background_thread.daemon = True
        # background_thread.start()

        worker_thread = threading.Thread(target=worker, args=(worker_time, ))
        worker_thread.daemon = True
        worker_thread.start()
        time.sleep(10)

        attempts = 0
        server = None
        while True:
            if attempts > 30:
                raise Exception('not able to start mining pool server')
            try:
                server = ThreadedTCPServer((_PoolHost, _PoolPort), TCPHandler)
            except Exception as e:
                lg.warn(e)
                time.sleep(10)
                attempts += 1
                continue
            break
        server_ip, server_port = server.server_address
        server_thread = threading.Thread(target=server.serve_forever)
        server_thread.daemon = True
        server_thread.start()

        if _Debug:
            lg.dbg(_DebugLevel, 'started mining pool server at %s:%d' % (server_ip, server_port))

        # print('Server thread is ready')

        # reactor.callFromThread(paydb_single)  # @UndefinedVariable

        reactor.callFromThread(starting_defer.callback, True)  # @UndefinedVariable

        server_thread.join()

        if _Debug:
            lg.dbg(_DebugLevel, 'mining pool server finished')

        server.shutdown()
        server.server_close()

    except Exception as e:
        lg.exc()
        reactor.callFromThread(starting_defer.errback, e)  # @UndefinedVariable


def percentage(percent, whole):
    return int((percent*whole)/100)


def read_mempool():
    mempool_db = sqlite3.connect(mempool_db_path)
    try:
        mempool_db.text_factory = str
        mp = mempool_db.cursor()
        mp.execute('SELECT * FROM transactions ORDER BY amount DESC')
        present = mp.fetchall()
    except:
        traceback.print_exc()
        present = None
    try:
        mempool_db.close()
    except:
        traceback.print_exc()
    return present


def payout(payout_threshold, myfee, othfee):
    global node_ip
    global node_port

    # print('Minimum payout is {} Bismuth'.format(str(payout_threshold)))
    # print('Current pool fee is {} Percent'.format(str(myfee)))
    # did_payout = False

    shares = sqlite3.connect(shares_db_path)
    shares.text_factory = str
    s = shares.cursor()

    #get sum of all shares not paid
    s.execute('SELECT sum(shares) FROM shares WHERE paid != 1')
    shares_total = s.fetchone()[0]
    #get sum of all shares not paid

    #get block threshold
    try:
        s.execute('SELECT min(timestamp) FROM shares WHERE paid != 1')
        block_threshold = float(s.fetchone()[0])
    except:
        block_threshold = time.time()
    #get block threshold
    # print('Pool: block_threshold', block_threshold, 'shares_total', shares_total, 'address', address)

    #get eligible blocks
    # conn = sqlite3.connect(ledger_path_conf)
    # conn.text_factory = str
    # c = conn.cursor()
    # reward_list = []
    # for row in c.execute('SELECT * FROM transactions WHERE address = ? AND CAST(timestamp AS INTEGER) >= ? AND reward != 0', (address, ) + (block_threshold, )):
    #     reward_list.append(float(row[9]))
    # c.close()

    t = socks.socksocket()
    t.connect((node_ip, int(node_port)))  # connect to local node
    connections.send(t, 'listreward', 10)
    connections.send(t, address, 10)
    connections.send(t, str(block_threshold), 10)
    listreward_reply = connections.receive(t, 10)
    t.close()
    reward_list = []
    for row in listreward_reply:
        reward_list.append(float(row[9]))

    # print('reward_list', reward_list)

    super_total = sum(reward_list)
    # print('Pool: super_total', super_total, 'reward_list', reward_list)
    #get eligible blocks

    # so now we have sum of shares, total reward, block threshold

    # reduce total rewards by total fees percentage
    reward_total = '%.8f' % (((100 - (myfee + othfee))*super_total)/100)
    reward_total = float(reward_total)

    print('Pool: reward_total=%r' % reward_total)

    if reward_total > 0:

        # calculate alt address fee

        ft = super_total - reward_total
        try:
            at = '%.8f' % (ft*(othfee/(myfee + othfee)))
        except:
            at = 0

        # calculate reward per share
        reward_per_share = reward_total/shares_total

        # calculate shares threshold for payment

        shares_threshold = math.floor(payout_threshold/reward_per_share)

        # print('reward_per_share1', reward_per_share, 'shares_threshold', shares_threshold)
        #get unique addresses
        addresses = []
        for row in s.execute('SELECT * FROM shares'):
            shares_address = row[0]

            if shares_address not in addresses:
                addresses.append(shares_address)
        # print('payout address', addresses)
        #get unique addresses

        # prepare payout address list with number of shares and new total shares
        payadd = []
        new_sum = 0
        for x in addresses:
            s.execute('SELECT sum(shares) FROM shares WHERE address = ? AND paid != 1', (x, ))
            shares_sum = s.fetchone()[0]

            if shares_sum == None:
                shares_sum = 0
            print(x, shares_sum, shares_threshold)
            if shares_sum > shares_threshold:
                payadd.append([x, shares_sum])
                new_sum = new_sum + shares_sum
        #prepare payout address list with number of shares and new total shares
        # print('payadd', payadd)

        # recalculate reward per share now we have removed those below payout threshold
        try:
            reward_per_share = reward_total/new_sum

        except:
            reward_per_share = 0

        # print('reward_per_share2', reward_per_share)

        paylist = []
        for p in payadd:
            payme = '%.8f' % (p[1]*reward_per_share)
            paylist.append([p[0], payme])

        if othfee > 0:
            paylist.append([alt_add, at])

        # print('paylist', paylist)
        payout_passed = 0
        for r in paylist:
            # print(r)
            recipient = r[0]
            claim = float(r[1])

            payout_passed = 1
            openfield = 'pool'
            keep = 0
            fee = 0
            # fee = float('%.8f' % float(0.01 + (float(len(openfield))/100000) + (keep)))  # 0.01 + openfield fee + keep fee
            #make payout

            timestamp = '%.2f' % time.time()
            transaction = (str(timestamp), str(address), str(recipient), '%.8f' % float(claim - fee), str(keep), str(openfield))  # this is signed
            # print transaction

            h = SHA.new(str(transaction).encode('utf-8'))
            signer = PKCS1_v1_5.new(key)
            signature = signer.sign(h)
            signature_enc = base64.b64encode(signature)
            # print('Encoded Signature: {}'.format(signature_enc.decode('utf-8')))

            verifier = PKCS1_v1_5.new(key)
            if verifier.verify(h, signature) == True:
                # print('The signature is valid, proceeding to send transaction')
                txid = signature_enc[:56]
                mytxid = txid.decode('utf-8')
                tx_submit = (str(timestamp), str(address), str(recipient), '%.8f' % float(claim - fee), str(signature_enc.decode('utf-8')), str(public_key_hashed.decode('utf-8')), str(keep), str(openfield))  #float kept for compatibility

                t = socks.socksocket()
                t.connect((node_ip, int(node_port)))  # connect to local node

                connections.send(t, 'mpinsert', 10)
                connections.send(t, [tx_submit], 10)
                reply = connections.receive(t, 10)
                t.close()
                # did_payout = True
                if _Debug:
                    lg.dbg(_DebugLevel, 'transaction {} sent with reply {}'.format(tx_submit, reply))
            else:
                # print('Pool: Invalid signature')
                reply = 'Invalid signature'

            s.execute('UPDATE shares SET paid = 1 WHERE address = ?', (recipient, ))
            shares.commit()

        if payout_passed == 1:
            s.execute('UPDATE shares SET timestamp = ?', (time.time(), ))
            shares.commit()

        # calculate payouts
        #payout

        # archive paid shares
        s.execute('SELECT * FROM shares WHERE paid = 1')
        pd = s.fetchall()

        if pd == None:
            pass
        else:
            archive = sqlite3.connect(archive_db_path)
            archive.text_factory = str
            a = archive.cursor()

            for sh in pd:
                a.execute('INSERT INTO shares VALUES (?,?,?,?,?,?,?,?)', (sh[0], sh[1], sh[2], sh[3], sh[4], sh[5], sh[6], sh[7]))

            archive.commit()
            a.close()
        # archive paid shares

    # clear nonces
    s.execute('DELETE FROM nonces')
    s.execute('DELETE FROM shares WHERE paid = 1')
    shares.commit()
    s.execute('VACUUM')
    #clear nonces
    s.close()

    # if did_payout:
    #     reactor.callFromThread(paydb_single, delay=10)  # @UndefinedVariable


def commit(cursor):
    # secure commit for slow nodes
    passed = 0
    while passed == 0:
        try:
            cursor.commit()
            passed = 1
        except Exception as e:
            lg.exc()
            # print('Retrying database execute due to ' + str(e))
            time.sleep(random.random())
            # secure commit for slow nodes


def execute(cursor, what):
    # secure execute for slow nodes
    passed = 0
    while passed == 0:
        try:
            # print cursor
            # print what

            cursor.execute(what)
            passed = 1
        except Exception as e:
            lg.exc()
            time.sleep(random.random())
            # secure execute for slow nodes
    return cursor


def execute_param(cursor, what, param):
    # secure execute for slow nodes
    passed = 0
    while passed == 0:
        try:
            # print cursor
            # print what
            cursor.execute(what, param)
            passed = 1
        except Exception as e:
            lg.exc()
            time.sleep(0.1)
            # secure execute for slow nodes
    return cursor


bin_format_dict = dict((x, format(ord(x), '8b').replace(' ', '0')) for x in '0123456789abcdef')


def bin_convert(string):
    return ''.join(bin_format_dict[x] for x in string)


def bin_convert_orig(string):
    return ''.join(format(ord(x), '8b').replace(' ', '0') for x in string)


def s_test(testString):
    if testString.isalnum():
        if (re.search('[abcdef]', testString)):
            if len(testString) == 56:
                return True
    else:
        return False


def n_test(testString):
    if testString.isalnum():
        if (re.search('[abcdef]', testString)):
            if len(testString) < 129:
                return True
    else:
        return False


def paydb():
    global new_time

    # disabled
    return True

    time.sleep(30)
    while True:
        # time.sleep(3601)
        # time.sleep(60 * 5)
        # v = float('%.2f' % time.time())
        # v1 = new_time
        # v2 = v - v1
        # if v2 < 100000:
        # print('Payout running...')
        payout(min_payout, pool_fee, alt_fee)
        # else:
        # print('Node over 1 mins out: %r ...payout delayed' % v2)
        time.sleep(60*5)


# def paydb_single(delay=0):
#     # print('Pool: paydb_single')
#     time.sleep(delay)
#     # reactor.callLater(delay, payout, min_payout, pool_fee, alt_fee)  # @UndefinedVariable
#     background_paydb_single_thread = threading.Thread(target=payout, args=(min_payout, pool_fee, alt_fee))
#     background_paydb_single_thread.daemon = True
#     background_paydb_single_thread.start()


def worker(s_time):
    global new_diff
    global new_hash
    global new_time
    global node_ip
    global node_port

    if _Debug:
        lg.dbg(_DebugLevel, 'about to connect to node at {}:{} from {}'.format(node_ip, node_port, threading.current_thread()))

    while True:

        time.sleep(s_time)

        try:

            n = socks.socksocket()
            n.connect((node_ip, int(node_port)))  # connect to local node

            connections.send(n, 'blocklast', 10)
            blocklast = connections.receive(n, 10)

            connections.send(n, 'diffget', 10)
            diff = connections.receive(n, 10)

            new_hash = blocklast[7]
            new_time = blocklast[1]
            new_diff = math.floor(diff[1])

            n.close()
            # print('Pool: difficulty={} blockhash={}'.format(str(new_diff), str(new_hash)))

        except Exception as e:
            traceback.print_exc()

        finally:
            try:
                n.close()
            except:
                traceback.print_exc()


class TCPHandler(socketserver.BaseRequestHandler):

    def handle(self):
        global new_diff
        global node_ip
        global node_port
        global _DataDirPath

        block_submitted = False
        share_added = False

        key = RSA.importKey(private_key_readable)

        self.allow_reuse_address = True

        peer_ip, peer_port = self.request.getpeername()

        try:
            try:
                data = connections.receive(self.request, 10)
            except:
                data = ''
                lg.warn('failed reading data from %r: %r' % (self.client_address, self.request))

            # if _Debug:
            #     lg.args(_DebugLevel, data=data, peer_ip=peer_ip, peer_port=peer_port)
            # print('Pool: received {} from {}:{}'.format(data, peer_ip, peer_port))  # will add custom ports later

            if data == 'getwork':  # sends the miner the blockhash and mining diff for shares

                result = read_mempool()
                # m = socks.socksocket()
                # m.connect((node_ip, int(node_port)))  # connect to local node
                # connections.send(m, 'api_mempool', 10)
                # result = connections.receive(m, 10)
                # m.close()

                work_send = []
                work_send.append((len(result), new_hash, mdiff, address, mdiff))

                connections.send(self.request, work_send, 10)
                # print('Work package sent.... {}'.format(work_send))

            elif data == 'block':  # from miner to node

                # sock
                #s1 = socks.socksocket()
                #if tor_conf == 1:
                #    s1.setproxy(socks.PROXY_TYPE_SOCKS5, "127.0.0.1", 9050)
                #s1.connect(("127.0.0.1", int(port)))  # connect to local node,
                # sock

                # receive nonce from miner
                miner_address = connections.receive(self.request, 10)
                # print('miner_address', miner_address)

                if not s_test(miner_address):
                    lg.err(_DebugLevel, 'bad miner address detected - changing to default')
                    miner_address = alt_add
                    #s1.close()

                else:

                    if _Debug:
                        lg.dbg(_DebugLevel, 'received a solution from miner {} ({})'.format(peer_ip, miner_address))

                    block_nonce = connections.receive(self.request, 10)
                    block_timestamp = (block_nonce[-1][0])
                    nonce = (block_nonce[-1][1])
                    mine_hash = ((block_nonce[-1][2]))  # block hash claimed
                    ndiff = ((block_nonce[-1][3]))  # network diff when mined
                    sdiffs = ((block_nonce[-1][4]))  # actual diff mined
                    mrate = ((block_nonce[-1][5]))  # total hash rate in khs
                    bname = ((block_nonce[-1][6]))  # base worker name
                    wnum = ((block_nonce[-1][7]))  # workers
                    wstr = ((block_nonce[-1][8]))  # worker number
                    wname = '{}{}'.format(bname, wstr)  # worker name

                    # print('Pool: Mined nonce details: {}'.format(block_nonce))
                    # print('Pool: Claimed hash: {}'.format(mine_hash))
                    # print('Pool: Claimed diff: {}'.format(sdiffs))

                    if not n_test(nonce):
                        lg.err('bad nonce format detected - closing connection')
                        self.close()
                    # print('Pool: Processing nonce.....')

                    diff = new_diff
                    db_block_hash = mine_hash

                    heavy3_path = os.path.join(_DataDirPath, 'heavy3a.bin')
                    mining_heavy3.mining_open(heavy3_path)
                    real_diff = mining_heavy3.diffme_heavy3(address, nonce, db_block_hash)
                    # mining_heavy3.mining_close()
                    """
                    mining_hash = bin_convert_orig(hashlib.sha224((address + nonce + db_block_hash).encode("utf-8")).hexdigest())
                    mining_condition = bin_convert_orig(db_block_hash)[0:diff]

                    if mining_condition in mining_hash:
                    """
                    # print('Pool: Solution has {} difficulty, current difficulty is {}'.format(real_diff, diff))
                    if real_diff >= int(diff):

                        # print('Pool: Difficulty requirement satisfied for mining')
                        # print('Pool: Sending block to nodes')

                        # cn = options.Get()
                        # cn.read()
                        # cport = cn.port
                        # try:
                        #     cnode_ip_conf = cn.node_ip_conf
                        # except:
                        #     cnode_ip_conf = cn.node_ip

                        #ctor_conf = cn.tor_conf
                        # cversion = cn.version

                        # print('Pool: Local node ip {} on port {}'.format(node_ip, node_port))

                        result = read_mempool()
                        # m = socks.socksocket()
                        # m.connect((node_ip, int(node_port)))  # connect to local node
                        # connections.send(m, 'api_mempool', 10)
                        # result = connections.receive(m, 10)
                        # print('I have got to receive mempool')
                        # m.close()

                        # include data
                        block_send = []
                        del block_send[:]  # empty
                        removal_signature = []
                        del removal_signature[:]  # empty

                        for dbdata in result:
                            transaction = (str(dbdata[0]), str(dbdata[1][:56]), str(dbdata[2][:56]), '%.8f' % float(dbdata[3]), str(dbdata[4]), str(dbdata[5]), str(dbdata[6]), str(dbdata[7]))  # create tuple
                            block_send.append(transaction)  # append tuple to list for each run
                            removal_signature.append(str(dbdata[4]))  # for removal after successful mining

                        # claim reward
                        transaction_reward = (str(block_timestamp), str(address[:56]), str(address[:56]), '%.8f' % float(0), '0', str(nonce))  # only this part is signed!
                        # print('transaction_reward', transaction_reward)

                        h = SHA.new(str(transaction_reward).encode('utf-8'))
                        signer = PKCS1_v1_5.new(key)
                        signature = signer.sign(h)
                        signature_enc = base64.b64encode(signature)

                        if signer.verify(h, signature) == True:
                            # print('Pool: Signature valid')

                            block_send.append((str(block_timestamp), str(address[:56]), str(address[:56]), '%.8f' % float(0), str(signature_enc.decode('utf-8')), str(public_key_hashed.decode('utf-8')), '0', str(nonce)))  # mining reward tx
                            # print('Pool: Block to send: {}'.format(block_send))

                            if not any(isinstance(el, list) for el in block_send):  # if it's not a list of lists (only the mining tx and no others)
                                new_list = []
                                new_list.append(block_send)
                                block_send = new_list  # make it a list of lists
                                # print(block_send)

                        peer_dict = {}
                        # peer_dict[node_ip] = node_port

                        # if True:
                        with open(m_peer_file) as f:
                            peer_dict = json.load(f)

                            if '127.0.0.1' in peer_dict:
                                peer_dict = {'127.0.0.1': peer_dict['127.0.0.1']}

                            for k, v in peer_dict.items():
                                peer_ip = k
                                # app_log.info(HOST)
                                peer_port = int(v)
                                # app_log.info(PORT)
                                # connect to all nodes
                                # print('Pool: Proceeding to submit mined block to {}:{}'.format(peer_ip, peer_port))
                                try:
                                    s = socks.socksocket()
                                    s.settimeout(0.3)
                                    #if ctor_conf == 1:
                                    #    s.setproxy(socks.PROXY_TYPE_SOCKS5, "127.0.0.1", 9050)
                                    s.connect((peer_ip, int(peer_port)))  # connect to node in peerlist

                                    connections.send(s, 'block', 10)
                                    #connections.send(s, address, 10)
                                    connections.send(s, block_send, 10)
                                    block_submitted = True
                                    if _Debug:
                                        lg.dbg(_DebugLevel, 'block submitted to {}:{}'.format(peer_ip, peer_port))
                                except Exception as e:
                                    lg.exc()
                                    # print('Pool: Could not submit block to {}:{} because {}'.format(peer_ip, peer_port, e))
                                finally:
                                    s.close()

                    if diff < mdiff:
                        diff_shares = diff
                    else:
                        diff_shares = mdiff

                    shares = sqlite3.connect(shares_db_path)
                    shares.text_factory = str
                    s_cur = shares.cursor()

                    # protect against used share resubmission
                    execute_param(s_cur, ('SELECT nonce FROM nonces WHERE nonce = ?'), (nonce, ))

                    try:
                        result = s_cur.fetchone()[0]
                        lg.err('Pool: Miner trying to reuse a share, ignored')

                    except:
                        # protect against used share resubmission
                        """
                        mining_condition = bin_convert_orig(db_block_hash)[0:diff_shares] #floor set by pool
                        if mining_condition in mining_hash:
                        """
                        if real_diff >= diff_shares and address != miner_address:
                            # print('Pool: Difficulty requirement satisfied for saving shares')

                            execute_param(s, ('INSERT INTO nonces VALUES (?)'), (nonce, ))
                            commit(shares)

                            timestamp = '%.2f' % time.time()

                            s_cur.execute('INSERT INTO shares VALUES (?,?,?,?,?,?,?,?)', (str(miner_address), str(1), timestamp, '0', str(mrate), bname, str(wnum), wname))
                            shares.commit()
                            share_added = True

                            if _Debug:
                                lg.dbg(_DebugLevel, 'added new share for %r' % miner_address)

                        # else:
                        # print('Pool: Difficulty requirement not satisfied for anything')

                    s_cur.close()

            self.request.close()
        except Exception as e:
            lg.exc()
            # print('Pool: Error: {}'.format(e))

        # if block_submitted and share_added:
        # background_paydb_single_thread = threading.Thread(target=paydb_single)
        # background_paydb_single_thread.daemon = True
        # background_paydb_single_thread.start()
        # reactor.callFromThread(paydb_single, delay=1)  # @UndefinedVariable

        # if _Debug:
        #     lg.args(_DebugLevel, block_submitted=block_submitted, share_added=share_added)

        # else:
        # print('Pool: block_submitted=%r share_added=%r' % (block_submitted, share_added))

    # print('Pool: Starting up...')


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass



from decimal import Decimal

DECIMAL_ZERO_2DP = Decimal('0.00')
DECIMAL_ZERO_8DP = Decimal('0.00000000')
DECIMAL_ZERO_10DP = Decimal('0.0000000000')


def quantize_two(value):
    if not value:
        return DECIMAL_ZERO_2DP
    value = Decimal(value)
    value = value.quantize(DECIMAL_ZERO_2DP)
    return value


def quantize_eight(value):
    if not value:
        # Will match 0 as well as False and None
        return DECIMAL_ZERO_8DP
    value = Decimal(value)
    value = value.quantize(DECIMAL_ZERO_8DP)
    return value


def quantize_ten(value):
    if not value:
        return DECIMAL_ZERO_10DP
    value = Decimal(value)
    value = value.quantize(DECIMAL_ZERO_10DP)
    return value



#!/usr/bin/env python
# udp_file_queue.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (udp_file_queue.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
"""
..module:: udp_file_queue.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from io import open
from io import StringIO

#------------------------------------------------------------------------------

import os
import time
import struct
import random

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng
from bitdust.lib import udp

from bitdust.system import tmpfile

from bitdust.contacts import contactsdb

#------------------------------------------------------------------------------

_Debug = False

#------------------------------------------------------------------------------

MAX_SIMULTANEOUS_STREAMS_PER_SESSION = 16
NUMBER_OF_STREAMS_TO_REMEMBER = MAX_SIMULTANEOUS_STREAMS_PER_SESSION*8*2

#------------------------------------------------------------------------------

_StreamCounter = 0

#------------------------------------------------------------------------------


class FileQueue:
    def __init__(self, session):
        self.session = session
        self.streams = {}
        self.outboxFiles = {}
        self.inboxFiles = {}
        self.outboxQueue = []
        self.dead_streams = []

    def make_unique_stream_id(self):
        global _StreamCounter
        _StreamCounter += 1
        n = 10 + max(contactsdb.contact_position(self.session.peer_idurl), 0) % 89
        return n*1000000 + random.randint(10, 99)*10000 + _StreamCounter % 10000

    def report_failed_outbox_queue(self, error_message):
        for filename, description, result_defer, keep_alive in self.outboxQueue:
            self.on_failed_outbox_queue_item(filename, description, error_message, result_defer, keep_alive)

    def report_failed_outbox_files(self, error_message):
        for outfile in self.outboxFiles.values():
            outfile.status = 'failed'
            outfile.error_message = error_message
            self.report_outbox_file(outfile)

    def report_failed_inbox_files(self, error_message):
        for infile in self.inboxFiles.values():
            infile.status = 'failed'
            infile.error_message = error_message
            self.report_inbox_file(infile)

    def close(self):
        for stream in self.streams.values():
            stream.on_close()
        self.outboxQueue = []

    def do_send_data(self, stream_id, outfile, output):
        #         if _Debug:
        #             import random
        #             if random.randint(1, 100) > 90:
        #                 return True
        newoutput = ''.join((struct.pack('i', stream_id), struct.pack('i', outfile.size), output))
        return self.session.send_packet(udp.CMD_DATA, strng.to_bin(newoutput))

    def do_send_ack(self, stream_id, infile, ack_data):
        #         if _Debug:
        #             import random
        #             if random.randint(1, 100) > 90:
        #                 return True
        newoutput = ''.join((struct.pack('i', stream_id), ack_data))
        return self.session.send_packet(udp.CMD_ACK, strng.to_bin(newoutput))

    def append_outbox_file(self, filename, description='', result_defer=None, keep_alive=True):
        from bitdust.transport.udp import udp_session
        self.outboxQueue.append((filename, description, result_defer, keep_alive))
        if _Debug:
            lg.out(18, 'udp_file_queue.append_outbox_file %s for %s : %s, streams=%d, queue=%d' % (os.path.basename(filename), self.session.peer_id, description, len(self.streams), len(self.outboxQueue)))
        udp_session.process_sessions([
            self.session,
        ])

    def insert_outbox_file(self, filename, description='', result_defer=None, keep_alive=True):
        from bitdust.transport.udp import udp_session
        self.outboxQueue.insert(0, (filename, description, result_defer, keep_alive))
        if _Debug:
            lg.out(18, 'udp_file_queue.insert_outbox_file %s for %s : %s, streams=%d, queue=%d' % (os.path.basename(filename), self.session.peer_id, description, len(self.streams), len(self.outboxQueue)))
        udp_session.process_sessions([
            self.session,
        ])

    def process_outbox_queue(self):
        has_reads = False
        while len(self.outboxQueue) > 0:
            if len(self.streams) >= MAX_SIMULTANEOUS_STREAMS_PER_SESSION:
                # lg.warn('too much streams: %d' % len(self.streams))
                break
            filename, description, result_defer, keep_alive = self.outboxQueue.pop(0)
            has_reads = True
            # we have a queue of files to be sent
            # somehow file may be removed before we start sending it
            # so I check it here and skip not existed files
            if not os.path.isfile(filename):
                self.on_failed_outbox_queue_item(filename, description, 'file not exist', result_defer, keep_alive)
                continue
            try:
                filesize = os.path.getsize(filename)
            except:
                self.on_failed_outbox_queue_item(filename, description, 'can not get file size', result_defer, keep_alive)
                continue
            self.start_outbox_file(filename, filesize, description, result_defer, keep_alive)
        return has_reads

    def process_outbox_files(self):
        has_sends = False
        for outfile in self.outboxFiles.values():
            has_sends = has_sends or outfile.process()
        return has_sends

    def start_outbox_file(self, filename, filesize, description, result_defer, keep_alive):
        from bitdust.transport.udp import udp_interface
        from bitdust.transport.udp import udp_stream
        stream_id = self.make_unique_stream_id()
        if _Debug:
            lg.out(12, 'udp_file_queue.start_outbox_file %d %s %s %d %s' % (stream_id, description, os.path.basename(filename), filesize, self.session.peer_id))
        self.outboxFiles[stream_id] = OutboxFile(self, stream_id, filename, filesize, description, result_defer, keep_alive)
        self.streams[stream_id] = udp_stream.create(stream_id, self.outboxFiles[stream_id], self)
        if keep_alive:
            d = udp_interface.interface_register_file_sending(self.session.peer_id, self.session.peer_idurl, filename, description)
            d.addCallback(self.on_outbox_file_registered, stream_id)
            d.addErrback(self.on_outbox_file_register_failed, stream_id)
            self.outboxFiles[stream_id].registration = d

    def start_inbox_file(self, stream_id, data_size):
        from bitdust.transport.udp import udp_interface
        from bitdust.transport.udp import udp_stream
        if _Debug:
            lg.out(12, 'udp_file_queue.start_inbox_file %d %d %s' % (stream_id, data_size, self.session.peer_id))
        self.inboxFiles[stream_id] = InboxFile(self, stream_id, data_size)
        self.streams[stream_id] = udp_stream.create(stream_id, self.inboxFiles[stream_id], self)
        d = udp_interface.interface_register_file_receiving(self.session.peer_id, self.session.peer_idurl, self.inboxFiles[stream_id].filename, self.inboxFiles[stream_id].size)
        d.addCallback(self.on_inbox_file_registered, stream_id)
        d.addErrback(self.on_inbox_file_register_failed, stream_id)
        self.inboxFiles[stream_id].registration = d

    def erase_stream(self, stream_id):
        del self.streams[stream_id]
        self.dead_streams.append(stream_id)
        if len(self.dead_streams) > NUMBER_OF_STREAMS_TO_REMEMBER:
            self.dead_streams.pop(0)
        if _Debug:
            lg.out(18, 'udp_file_queue.erase_stream %s' % stream_id)

    def close_outbox_file(self, stream_id):
        # if _Debug:
        #     lg.out(18, 'close_outbox_file %r' % stream_id)
        if self.outboxFiles.get(stream_id):
            self.outboxFiles[stream_id].close()
            del self.outboxFiles[stream_id]
        else:
            lg.warn('outgoing UDP file %s not exist' % stream_id)

    def close_inbox_file(self, stream_id):
        # if _Debug:
        #     lg.out(18, 'close_inbox_file %r' % stream_id)
        if self.inboxFiles.get(stream_id):
            self.inboxFiles[stream_id].close()
            del self.inboxFiles[stream_id]
        else:
            lg.warn('incoming UDP file %s not exist' % stream_id)

    def report_outbox_file(self, outfile):
        from bitdust.transport.udp import udp_interface
        if _Debug:
            lg.out(18, 'udp_file_queue.report_outbox_file %s %s %d bytes "%s"' % (outfile.transfer_id, outfile.status, outfile.bytes_delivered, outfile.error_message))
        udp_interface.interface_unregister_file_sending(outfile.transfer_id, outfile.status, outfile.bytes_delivered, outfile.error_message)

    def report_inbox_file(self, infile):
        from bitdust.transport.udp import udp_interface
        if _Debug:
            lg.out(18, 'udp_file_queue.report_inbox_file {%s} %s %s %d bytes "%s"' % (os.path.basename(infile.filename), infile.transfer_id, infile.status, infile.bytes_received, infile.error_message))
        udp_interface.interface_unregister_file_receiving(infile.transfer_id, infile.status, infile.bytes_received, infile.error_message)

    #-------------------------------------------------------------------------

    def on_received_data_packet(self, payload):
        inp = StringIO(payload)
        try:
            stream_id = int(struct.unpack('i', inp.read(4))[0])
            data_size = int(struct.unpack('i', inp.read(4))[0])
        except:
            inp.close()
            lg.exc()
            return
        if not self.session.peer_id:
            inp.close()
            if _Debug:
                lg.warn('SEND ZERO ACK, peer id is unknown yet %s' % stream_id)
            self.do_send_ack(stream_id, None, '')
            return
        if stream_id not in list(self.streams.keys()):
            if stream_id in self.dead_streams:
                inp.close()
                # if _Debug:
                # lg.warn('SEND ZERO ACK, got old block %s' % stream_id)
                self.do_send_ack(stream_id, None, '')
                return
            if len(self.streams) >= 2*MAX_SIMULTANEOUS_STREAMS_PER_SESSION:
                # too many incoming streams, seems remote side is cheating - drop that session!
                # TODO: need to add some protection - keep a list of bad guys?
                inp.close()
                # lg.warn('too many incoming files for session %s' % str(self.session))
                # self.session.automat('shutdown')
                if _Debug:
                    lg.warn('SEND ZERO ACK, too many active streams: %d  skipped: %s %s' % (len(self.streams), stream_id, self.session.peer_id))
                self.do_send_ack(stream_id, None, '')
                return
            self.start_inbox_file(stream_id, data_size)
        try:
            self.streams[stream_id].on_block_received(inp)
        except:
            lg.exc()
        inp.close()

    def on_received_ack_packet(self, payload):
        inp = StringIO(payload)
        try:
            stream_id = int(struct.unpack('i', inp.read(4))[0])
        except:
            inp.close()
            lg.exc()
            # self.session.automat('shutdown')
            return
        if stream_id not in list(self.streams.keys()):
            inp.close()
            # if not self.receivedFiles.has_key(stream_id):
            # lg.warn('unknown stream_id=%d in ACK packet from %s' % (
            #     stream_id, self.session.peer_address))
            # self.session.automat('shutdown')
            if stream_id in self.dead_streams:
                # print 'old ack', stream_id
                pass
            else:
                if _Debug:
                    lg.warn('%s - what a stream ???' % stream_id)
            # self.session.automat('shutdown')
            return
        try:
            self.streams[stream_id].on_ack_received(inp)
        except:
            lg.exc()
            self.session.automat('shutdown')
        inp.close()

    def on_inbox_file_done(self, stream_id):
        assert stream_id in list(self.inboxFiles.keys())
        infile = self.inboxFiles[stream_id]
        if _Debug:
            lg.out(18, 'udp_file_queue.on_inbox_file_done %s (%d bytes) %s "%s" registration=%r' % (stream_id, infile.size, infile.status, infile.error_message, infile.registration))
        if infile.registration:
            return
        if infile.transfer_id:
            self.report_inbox_file(infile)
        s = self.streams[stream_id]
        s.on_close()
        s = None

    def on_outbox_file_done(self, stream_id):
        assert stream_id in list(self.outboxFiles.keys())
        outfile = self.outboxFiles[stream_id]
        if _Debug:
            lg.out(18, 'udp_file_queue.on_outbox_file_done %s (%d bytes) %s "%s" registration=%r' % (stream_id, outfile.size, outfile.status, outfile.error_message, outfile.registration))
        if outfile.registration:
            return
        if outfile.transfer_id:
            self.report_outbox_file(outfile)
        if outfile.result_defer:
            outfile.result_defer.callback((outfile, outfile.status, outfile.error_message))
            outfile.result_defer = None
        s = self.streams[stream_id]
        s.on_close()
        s = None
        reactor.callLater(0, self.process_outbox_queue)

    def on_inbox_file_registered(self, response, stream_id):
        if stream_id not in list(self.inboxFiles.keys()):
            lg.warn('stream %d not found in the inboxFiles' % stream_id)
            return
        try:
            transfer_id = int(response)
        except:
            transfer_id = None
        infile = self.inboxFiles[stream_id]
        infile.transfer_id = transfer_id
        infile.registration = None
        if _Debug:
            lg.out(18, 'udp_file_queue.on_inbox_file_registered %d transfer_id=%r status=%s' % (stream_id, transfer_id, infile.status))
        if infile.status:
            self.report_inbox_file(infile)
            s = self.streams[stream_id]
            s.on_close()
            s = None

    def on_inbox_file_register_failed(self, err, stream_id):
        lg.out(2, 'udp_file_queue.on_inbox_file_register_failed ERROR failed to register, stream_id=%s, err: %s' % (str(stream_id), err))
        self.session.automat('shutdown')

    def on_outbox_file_registered(self, response, stream_id):
        if stream_id not in list(self.outboxFiles.keys()):
            lg.warn('stream %d not found in the outboxFiles' % stream_id)
            return
        try:
            transfer_id = int(response)
        except:
            transfer_id = None
        outfile = self.outboxFiles[stream_id]
        outfile.transfer_id = transfer_id
        outfile.registration = None
        if _Debug:
            lg.out(18, 'udp_file_queue.on_outbox_file_registered %d transfer_id=%r status=%s' % (stream_id, transfer_id, outfile.status))
        if outfile.status:
            self.report_outbox_file(outfile)
            s = self.streams[stream_id]
            s.on_close()
            s = None

    def on_outbox_file_register_failed(self, err, stream_id):
        lg.out(2, 'udp_file_queue.on_outbox_file_register_failed ERROR failed to register, stream_id=%s :\n%s' % (str(stream_id), str(err)))
        lg.out(6, 'udp_file_queue.on_outbox_file_register_failed close session %s' % self.session)
        self.session.automat('shutdown')

    def on_failed_outbox_queue_item(self, filename, description='', error_message='', result_defer=None, keep_alive=True):
        from bitdust.transport.udp import udp_interface
        if _Debug:
            lg.out(18, 'udp_file_queue.failed_outbox_queue_item %s because %s' % (filename, error_message))
        try:
            udp_interface.interface_cancelled_file_sending(self.session.peer_id, filename, 0, description, error_message).addErrback(lambda err: lg.exc(err))
        except Exception as exc:
            lg.warn(str(exc))
        if result_defer:
            result_defer.callback(((filename, description), 'failed', error_message))

    def on_timeout_receiving(self, stream_id):
        assert stream_id in list(self.inboxFiles.keys())
        infile = self.inboxFiles[stream_id]
        if _Debug:
            lg.out(18, 'udp_file_queue.on_timeout_receiving stream_id=%s %d : %s' % (stream_id, infile.bytes_received, infile.error_message))
        if infile.registration:
            return
        if infile.transfer_id:
            self.report_inbox_file(infile)
        if _Debug:
            lg.out(2, '!'*80)
        s = self.streams[stream_id]
        s.on_close()
        s = None

    def on_timeout_sending(self, stream_id):
        assert stream_id in list(self.outboxFiles.keys())
        outfile = self.outboxFiles[stream_id]
        if _Debug:
            lg.out(18, 'udp_file_queue.on_timeout_sending stream_id=%s %d/%d bytes sent : %s' % (stream_id, outfile.bytes_delivered, outfile.bytes_sent, outfile.error_message))
        if outfile.registration:
            return
        if outfile.transfer_id:
            self.report_outbox_file(outfile)
        if outfile.result_defer:
            outfile.result_defer.callback((outfile, outfile.status, outfile.error_message))
            outfile.result_defer = None
        if _Debug:
            lg.out(2, '!'*80)
        s = self.streams[stream_id]
        s.on_close()
        s = None

    def on_close_stream(self, stream_id):
        if _Debug:
            lg.out(18, 'udp_file_queue.on_close_stream %s' % stream_id)
        self.erase_stream(stream_id)

    def on_close_consumer(self, consumer):
        if isinstance(consumer, InboxFile):
            self.close_inbox_file(consumer.stream_id)
        elif isinstance(consumer, OutboxFile):
            self.close_outbox_file(consumer.stream_id)
        else:
            raise Exception('incorrect consumer object')


#------------------------------------------------------------------------------


class InboxFile():
    def __init__(self, queue, stream_id, size):
        """
        """
        self.typ = 'udp-in'
        self.transfer_id = None
        self.registration = None
        self.queue = queue
        self.stream_callback = None
        self.stream_id = stream_id
        self.fd, self.filename = tmpfile.make('udp-in', extension='.udp')
        self.size = size
        self.bytes_received = 0
        self.started = time.time()
        self.cancelled = False
        self.timeout = False
        self.status = None
        self.error_message = ''
        if _Debug:
            lg.out(18, 'udp_file_queue.InboxFile.__init__ {%s} [%d] from %s with %d bytes' % (os.path.basename(self.filename), self.stream_id, str(self.queue.session.peer_address), self.size))

    def __del__(self):
        if _Debug:
            lg.out(18, 'udp_file_queue.InboxFile.__del__ {%s} [%d]' % (os.path.basename(self.filename), self.stream_id))

    def set_stream_callback(self, stream_callback):
        self.stream_callback = stream_callback

    def clear_stream_callback(self):
        self.stream_callback = None

    def close(self):
        if _Debug:
            lg.out(18, 'udp_file_queue.InboxFile.close %d : %d received' % (self.stream_id, self.bytes_received))
        self.close_file()
        self.queue = None
        self.stream_callback = None

    def close_file(self):
        os.close(self.fd)
        self.fd = None

    def process(self, newdata):
        os.write(self.fd, newdata)
        self.bytes_received += len(newdata)

    def is_done(self):
        return self.bytes_received == self.size

    def is_timed_out(self):
        return self.timeout

    def on_received_raw_data(self, newdata):
        self.process(newdata)
        return self.is_done()


#------------------------------------------------------------------------------


class OutboxFile():
    def __init__(self, queue, stream_id, filename, size, description='', result_defer=None, keep_alive=True):
        """
        """
        self.typ = 'udp-out'
        self.transfer_id = None
        self.registration = None
        self.queue = queue
        self.stream_callback = None
        self.stream_id = stream_id
        self.filename = filename
        self.size = size
        self.description = description
        self.result_defer = result_defer
        self.keep_alive = keep_alive
        self.bytes_sent = 0
        self.bytes_delivered = 0
        self.buffer = b''
        self.eof = False
        self.cancelled = False
        self.timeout = False
        self.status = None
        self.error_message = ''
        self.started = time.time()
        self.fileobj = open(self.filename, 'rb')
        if _Debug:
            lg.out(18, 'udp_file_queue.OutboxFile.__init__ {%s} [%d] to %s with %d bytes' % (os.path.basename(self.filename), self.stream_id, str(self.queue.session.peer_address), self.size))

    def __del__(self):
        """
        """
        if _Debug:
            lg.out(18, 'udp_file_queue.OutboxFile.__del__ {%s} [%d] file:%r' % (os.path.basename(self.filename), self.stream_id, self.fileobj))

    def set_stream_callback(self, stream_callback):
        self.stream_callback = stream_callback

    def clear_stream_callback(self):
        self.stream_callback = None

    def close(self):
        if _Debug:
            lg.out(18, 'udp_file_queue.OutboxFile.close %s %d/%d' % (self.stream_id, self.bytes_sent, self.bytes_delivered))
        self.close_file()
        self.queue = None
        self.stream_callback = None
        self.buffer = b''
        self.description = None
        self.result_defer = None

    def close_file(self):
        self.fileobj.close()
        self.fileobj = None

    def is_eof(self):
        return self.eof

    def is_done(self):
        return self.eof and self.size == self.bytes_delivered

    def is_cancelled(self):
        return self.cancelled

    def is_timed_out(self):
        return self.timeout

    def count_size(self, more_bytes_delivered):
        self.bytes_delivered += more_bytes_delivered

    def cancel(self):
        if _Debug:
            lg.out(18, 'udp_file_queue.OutboxFile.cancel %s %d/%d' % (self.stream_id, self.bytes_sent, self.bytes_delivered))
        self.cancelled = True
        self.status = 'failed'
        self.error_message = 'transfer cancelled'

    def process(self):
        if self.eof:
            return False
        from bitdust.transport.udp import udp_stream
        has_sends = False
        while True:
            if not self.buffer:
                if not self.fileobj:
                    return False
                data = self.fileobj.read(udp_stream.CHUNK_SIZE)
                if not data:
                    if _Debug:
                        lg.out(18, 'udp_file_queue.OutboxFile.process reach EOF state %d' % self.stream_id)
                    self.eof = True
                    break
                self.buffer = data
            if not self.stream_callback:
                break
            try:
                self.stream_callback(self.buffer)
            except udp_stream.BufferOverflow:
                break
            self.bytes_sent += len(self.buffer)
            self.buffer = b''
            has_sends = True
        return has_sends

    def on_sent_raw_data(self, bytes_delivered):
        self.count_size(bytes_delivered)
        if self.is_done():
            return True
        if self.is_timed_out():
            return True
        if self.is_cancelled():
            return True
        self.process()
        return False



#!/usr/bin/python
# backup_rebuilder.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (backup_rebuilder.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: backup_rebuilder.

.. raw:: html

    <a href="https://bitdust.io/automats/backup_rebuilder/backup_rebuilder.png" target="_blank">
    <img src="https://bitdust.io/automats/backup_rebuilder/backup_rebuilder.png" style="max-width:100%;">
    </a>

This is a state machine to run the rebuilding process.
If some pieces is missing, need to reconstruct them asap.

To do that you need to know needed number of pieces on hands,
so need to request the missing segments.

The ``backup_rebuilder()`` machine works on backups one by one (keep them in queue)
and can be stopped and started at any time.

The whole process here may be stopped from ``backup_monitor()`` by
setting a flag in the ``isStopped()`` condition.
This is need to be able to stop the rebuilding process -
to do rebuilding of a single block we need start a blocking code.

EVENTS:
    * :red:`backup-ready`
    * :red:`found-missing`
    * :red:`inbox-data-packet`
    * :red:`init`
    * :red:`instant`
    * :red:`no-requests`
    * :red:`rebuilding-finished`
    * :red:`requests-sent`
    * :red:`start`
    * :red:`timer-1sec`
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 14

#------------------------------------------------------------------------------

import os
import sys

#------------------------------------------------------------------------------

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in backup_rebuilder.py')

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import packetid

from bitdust.automats import automat

from bitdust.system import bpio

from bitdust.main import settings

from bitdust.contacts import contactsdb

from bitdust.userid import my_id
from bitdust.userid import global_id
from bitdust.userid import id_url

from bitdust.raid import eccmap
from bitdust.raid import raid_worker

from bitdust.services import driver

#------------------------------------------------------------------------------

_BackupRebuilder = None
_StoppedFlag = True

_BackupIDsExclude = set()
_BackupIDsQueue = []
_BlockRebuildersQueue = []

#------------------------------------------------------------------------------


def A(event=None, *args, **kwargs):
    """
    Access method to interact with the state machine.
    """
    global _BackupRebuilder
    if _BackupRebuilder is None:
        _BackupRebuilder = BackupRebuilder(
            name='backup_rebuilder',
            state='STOPPED',
            debug_level=_DebugLevel,
            log_events=_Debug,
            log_transitions=_Debug,
        )
    if event is not None:
        _BackupRebuilder.automat(event, *args, **kwargs)
    return _BackupRebuilder


def Destroy():
    """
    Destroy backup_rebuilder() automat and remove its instance from memory.
    """
    global _BackupRebuilder
    if _BackupRebuilder is None:
        return
    _BackupRebuilder.destroy()
    del _BackupRebuilder
    _BackupRebuilder = None


class BackupRebuilder(automat.Automat):

    """
    A class to prepare and run rebuilding operations.
    """

    fast = False

    timers = {
        'timer-1sec': (1.0, ['REQUEST']),
    }

    def init(self):
        """
        Initialize needed variables.
        """
        self.currentBackupID = None  # currently working on this backup
        self.currentCustomerIDURL = None  # stored by this customer
        # list of missing blocks we work on for current backup
        self.workingBlocksQueue = []
        self.backupsWasRebuilt = []
        self.missingPackets = 0
        self.log_transitions = _Debug
        from bitdust.stream import data_sender
        data_sender.A().addStateChangedCallback(self._on_data_sender_state_changed)

    def shutdown(self):
        from bitdust.stream import data_sender
        data_sender.A().removeStateChangedCallback(self._on_data_sender_state_changed)

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        This method is called every time when my state is changed.

        Need to notify backup_monitor() machine about my new state.
        """
        # global_state.set_global_state('REBUILD ' + newstate)
        if newstate in [
            'NEXT_BACKUP',
            'REQUEST',
            'REBUILDING',
        ]:
            if event != 'instant':
                self.automat('instant')
            else:
                if oldstate == 'DONE':
                    self.automat('instant')
        elif newstate == 'DONE' or newstate == 'STOPPED':
            if driver.is_on('service_backups'):
                from bitdust.storage import backup_monitor
                backup_monitor.A('backup_rebuilder.state', newstate)

    def A(self, event, *args, **kwargs):
        #---STOPPED---
        if self.state == 'STOPPED':
            if event == 'init':
                self.doClearStoppedFlag(*args, **kwargs)
            elif event == 'start':
                self.state = 'NEXT_BACKUP'
                self.doClearStoppedFlag(*args, **kwargs)
        #---NEXT_BACKUP---
        elif self.state == 'NEXT_BACKUP':
            if event == 'instant' and not self.isMoreBackups(*args, **kwargs) and not self.isStopped(*args, **kwargs):
                self.state = 'DONE'
            elif event == 'instant' and self.isStopped(*args, **kwargs):
                self.state = 'STOPPED'
            elif event == 'instant' and not self.isStopped(*args, **kwargs) and self.isMoreBackups(*args, **kwargs):
                self.state = 'PREPARE'
                self.doOpenNextBackup(*args, **kwargs)
                self.doScanBrokenBlocks(*args, **kwargs)
        #---REQUEST---
        elif self.state == 'REQUEST':
            if (event == 'timer-1sec' or event == 'inbox-data-packet' or event == 'requests-sent' or event == 'found-missing') and self.isChanceToRebuild(*args, **kwargs) and not self.isStopped(*args, **kwargs):
                self.state = 'REBUILDING'
                self.doAttemptRebuild(*args, **kwargs)
            elif (
                (event == 'found-missing' and not self.isChanceToRebuild(*args, **kwargs)) or
                (event == 'no-requests' or ((event == 'timer-1sec' or event == 'inbox-data-packet') and self.isRequestQueueEmpty(*args, **kwargs) and not self.isMissingPackets(*args, **kwargs)))
            ) and not self.isStopped(*args, **kwargs):
                self.state = 'DONE'
                self.doCloseThisBackup(*args, **kwargs)
            elif ((event == 'instant' or event == 'timer-1sec' or event == 'requests-sent' or event == 'no-requests' or event == 'found-missing') and
                  self.isStopped(*args, **kwargs)) or (event == 'timer-1sec' and self.isRequestQueueEmpty(*args, **kwargs) and not self.isChanceToRebuild(*args, **kwargs) and not self.isStopped(*args, **kwargs)):
                self.state = 'STOPPED'
                self.doCloseThisBackup(*args, **kwargs)
        #---PREPARE---
        elif self.state == 'PREPARE':
            if event == 'backup-ready' and (not self.isMoreBackups(*args, **kwargs) and not self.isMoreBlocks(*args, **kwargs)):
                self.state = 'DONE'
                self.doCloseThisBackup(*args, **kwargs)
            elif (event == 'instant' or event == 'backup-ready') and self.isStopped(*args, **kwargs):
                self.state = 'STOPPED'
                self.doCloseThisBackup(*args, **kwargs)
            elif (event == 'instant' or event == 'backup-ready') and not self.isStopped(*args, **kwargs) and self.isMoreBlocks(*args, **kwargs):
                self.state = 'REQUEST'
                self.doRequestAvailablePieces(*args, **kwargs)
            elif event == 'backup-ready' and not self.isStopped(*args, **kwargs) and not self.isMoreBlocks(*args, **kwargs) and self.isMoreBackups(*args, **kwargs):
                self.state = 'NEXT_BACKUP'
                self.doCloseThisBackup(*args, **kwargs)
        #---REBUILDING---
        elif self.state == 'REBUILDING':
            if event == 'rebuilding-finished' and not self.isStopped(*args, **kwargs) and not self.isMoreBlocks(*args, **kwargs):
                self.state = 'PREPARE'
                self.doScanBrokenBlocks(*args, **kwargs)
            elif (event == 'instant' or event == 'rebuilding-finished') and self.isStopped(*args, **kwargs):
                self.state = 'STOPPED'
                self.doCloseThisBackup(*args, **kwargs)
                self.doKillRebuilders(*args, **kwargs)
            elif event == 'rebuilding-finished' and not self.isStopped(*args, **kwargs) and self.isMoreBlocks(*args, **kwargs):
                self.state = 'REQUEST'
                self.doRequestAvailablePieces(*args, **kwargs)
        #---DONE---
        elif self.state == 'DONE':
            if event == 'start' or (event == 'instant' and self.isSomeRebuilts(*args, **kwargs)):
                self.state = 'NEXT_BACKUP'
                self.doClearStoppedFlag(*args, **kwargs)
        return None

    def isSomeRebuilts(self, *args, **kwargs):
        """
        Condition method.
        """
        return len(self.backupsWasRebuilt) > 0

    def isMoreBackups(self, *args, **kwargs):
        """
        Condition method.
        """
        global _BackupIDsQueue
        global _BackupIDsExclude
        more_backups = set(_BackupIDsQueue)
        more_backups.difference_update(_BackupIDsExclude)
        return len(more_backups) > 0

    def isMoreBlocks(self, *args, **kwargs):
        """
        Condition method.
        """
        # because started from 0,  -1 means not found
        return len(self.workingBlocksQueue) > 0

    def isMissingPackets(self, *args, **kwargs):
        """
        Condition method.
        """
        return self.missingPackets > 0

    def isStopped(self, *args, **kwargs):
        """
        Condition method.
        """
        return ReadStoppedFlag()

    def isChanceToRebuild(self, *args, **kwargs):
        """
        Condition method.
        """
        from bitdust.storage import backup_matrix
        # start checking in reverse order, see below for explanation
        for blockIndex in range(len(self.workingBlocksQueue) - 1, -1, -1):
            blockNumber = self.workingBlocksQueue[blockIndex]
            if eccmap.Current().CanMakeProgress(backup_matrix.GetLocalDataArray(self.currentBackupID, blockNumber), backup_matrix.GetLocalParityArray(self.currentBackupID, blockNumber)):
                return True
        return False

    def isRequestQueueEmpty(self, *args, **kwargs):
        """
        Condition method.
        """
        from bitdust.stream import io_throttle
        for supplierNum in range(contactsdb.num_suppliers()):
            supplierID = contactsdb.supplier(supplierNum)
            if io_throttle.HasBackupIDInRequestQueue(supplierID, self.currentBackupID):
                return False
        return True

    def doOpenNextBackup(self, *args, **kwargs):
        """
        Action method.
        """
        global _BackupIDsQueue
        global _BackupIDsExclude
        more_backups = set(_BackupIDsQueue)
        more_backups.difference_update(_BackupIDsExclude)
        # check it, may be we already fixed all things
        if len(more_backups) == 0:
            self.automat('backup-ready')
            if _Debug:
                lg.out(_DebugLevel, 'backup_rebuilder.doOpenNextBackup SKIP, queue is empty')
            return
        # take a first backup from queue to work on it
        self.currentBackupID = list(more_backups)[0]
        # _BackupIDsQueue.pop(self.currentBackupID)
        self.currentCustomerIDURL = packetid.CustomerIDURL(self.currentBackupID)
        if _Debug:
            lg.out(_DebugLevel, 'backup_rebuilder.doOpenNextBackup %s started, queue length: %d' % (self.currentBackupID, len(_BackupIDsQueue)))

    def doCloseThisBackup(self, *args, **kwargs):
        """
        Action method.
        """
        global _BackupIDsQueue
        global _BackupIDsExclude
        self.workingBlocksQueue = []
        if _Debug:
            lg.out(_DebugLevel, 'backup_rebuilder.doCloseThisBackup %s about to finish, queue length: %d' % (self.currentBackupID, len(_BackupIDsQueue)))
        if self.currentBackupID:
            RemoveBackupToWork(self.currentBackupID)
            # clear requesting queue from previous task
            from bitdust.stream import io_throttle
            io_throttle.DeleteBackupRequests(self.currentBackupID)
        self.currentBackupID = None
        self.currentCustomerIDURL = None

    def doScanBrokenBlocks(self, *args, **kwargs):
        """
        Action method.
        """
        # if remote data structure is not exist for this backup - create it
        # this mean this is only local backup!
        from bitdust.storage import backup_matrix
        if self.currentBackupID not in backup_matrix.remote_files():
            backup_matrix.remote_files()[self.currentBackupID] = {}
            # we create empty remote info for every local block
            # range(0) should return []
            for blockNum in range(backup_matrix.local_max_block_numbers().get(self.currentBackupID, -1) + 1):
                backup_matrix.remote_files()[self.currentBackupID][blockNum] = {'D': [0]*contactsdb.num_suppliers(), 'P': [0]*contactsdb.num_suppliers()}
        # detect missing blocks from remote info
        self.workingBlocksQueue = backup_matrix.ScanMissingBlocks(self.currentBackupID)
        # find the correct max block number for this backup
        # we can have remote and local files
        # will take biggest block number from both
        backupMaxBlock = max(backup_matrix.remote_max_block_numbers().get(self.currentBackupID, -1), backup_matrix.local_max_block_numbers().get(self.currentBackupID, -1))
        # now need to remember this biggest block number
        # remote info may have less blocks - need to create empty info for
        # missing blocks
        for blockNum in range(backupMaxBlock + 1):
            if blockNum in backup_matrix.remote_files()[self.currentBackupID]:
                continue
            backup_matrix.remote_files()[self.currentBackupID][blockNum] = {'D': [0]*contactsdb.num_suppliers(), 'P': [0]*contactsdb.num_suppliers()}
        # clear requesting queue, remove old packets for this backup, we will
        # send them again
        from bitdust.stream import io_throttle
        io_throttle.DeleteBackupRequests(self.currentBackupID)
        if _Debug:
            lg.out(_DebugLevel, 'backup_rebuilder.doScanBrokenBlocks for %s : %s' % (self.currentBackupID, str(self.workingBlocksQueue)))
        self.automat('backup-ready')

    def doRequestAvailablePieces(self, *args, **kwargs):
        """
        Action method.
        """
        self._request_files()

    def doAttemptRebuild(self, *args, **kwargs):
        """
        Action method.
        """
        self.blocksSucceed = []
        if len(self.workingBlocksQueue) == 0:
            self.automat('rebuilding-finished')
            return
        # rebuild the backup blocks in reverse order, take last blocks first
        # in such way we can propagate information about how big is the whole backup as soon as possible!
        # remote machine can use simple formula [total size] = [file size] * [block number]
        # and calculate the whole size to be received
        self.blockIndex = len(self.workingBlocksQueue) - 1
        reactor.callLater(0, self._start_one_block)  # @UndefinedVariable

    def doKillRebuilders(self, *args, **kwargs):
        """
        Action method.
        """
        lg.warn('aborting raid worker for rebuilding %s' % self.currentBackupID)
        raid_worker.cancel_task('rebuild', self.currentBackupID)

    def doClearStoppedFlag(self, *args, **kwargs):
        """
        Action method.
        """
        self.backupsWasRebuilt = []
        ClearStoppedFlag()

    #-------------------------------------------------------------------------

    def _on_data_sender_state_changed(self, oldstate, newstate, event_string, *args, **kwargs):
        if oldstate == 'SCAN_BLOCKS' and newstate == 'READY':
            if event_string == 'scan-done' and args and isinstance(args[0], int):
                if args[0] > 0:
                    if _Debug:
                        lg.out(_DebugLevel, 'backup_rebuilder.BackupRebuilder._on_data_sender_state_changed is going to rebuild more')
                    self.automat('start')

    def _request_files(self):
        from bitdust.storage import backup_matrix
        from bitdust.stream import io_throttle
        from bitdust.stream import data_sender
        self.missingPackets = 0
        # here we want to request some packets before we start working to
        # rebuild the missed blocks
        availableSuppliers = backup_matrix.GetActiveArray(customer_idurl=self.currentCustomerIDURL)
        # remember how many requests we did on this iteration
        total_requests_count = 0
        # at the moment I do download everything I have available and needed
        if id_url.is_some_empty(contactsdb.suppliers(customer_idurl=self.currentCustomerIDURL)):
            if _Debug:
                lg.out(_DebugLevel, 'backup_rebuilder._request_files SKIP - empty supplier')
            self.automat('no-requests')
            return
        for supplierNum in range(contactsdb.num_suppliers(customer_idurl=self.currentCustomerIDURL)):
            supplierID = contactsdb.supplier(supplierNum, customer_idurl=self.currentCustomerIDURL)
            if not supplierID:
                continue
            requests_count = 0
            # we do requests in reverse order because we start rebuilding from
            # the last block
            for blockIndex in range(len(self.workingBlocksQueue) - 1, -1, -1):
                blockNum = self.workingBlocksQueue[blockIndex]
                # do not keep too many requests in the queue
                if io_throttle.GetRequestQueueLength(supplierID) >= 16:
                    break
                # also don't do too many requests at once
                if requests_count > 16:
                    break
                remoteData = backup_matrix.GetRemoteDataArray(self.currentBackupID, blockNum)
                remoteParity = backup_matrix.GetRemoteParityArray(self.currentBackupID, blockNum)
                localData = backup_matrix.GetLocalDataArray(self.currentBackupID, blockNum)
                localParity = backup_matrix.GetLocalParityArray(self.currentBackupID, blockNum)
                if supplierNum >= len(remoteData) or supplierNum >= len(remoteParity):
                    break
                if supplierNum >= len(localData) or supplierNum >= len(localParity):
                    break
                # if remote Data exist and is available because supplier is on-line,
                # but we do not have it on hand - do request
                if localData[supplierNum] == 0:
                    PacketID = packetid.MakePacketID(self.currentBackupID, blockNum, supplierNum, 'Data')
                    if remoteData[supplierNum] == 1:
                        if availableSuppliers[supplierNum]:
                            # if supplier is not alive - we can't request from him
                            if not io_throttle.HasPacketInRequestQueue(supplierID, PacketID):
                                customer, remotePath = packetid.SplitPacketID(PacketID)
                                filename = os.path.join(settings.getLocalBackupsDir(), customer, remotePath)
                                if not os.path.exists(filename):
                                    if io_throttle.QueueRequestFile(
                                        self._file_received,
                                        my_id.getIDURL(),
                                        PacketID,
                                        my_id.getIDURL(),
                                        supplierID,
                                    ):
                                        requests_count += 1
                    else:
                        # count this packet as missing
                        self.missingPackets += 1
                        # also mark this guy as one who dont have any data - nor local nor remote
                else:
                    # but if local Data already exists, but was not sent - do it now
                    if remoteData[supplierNum] != 1:
                        data_sender.A('new-data')
                # same for Parity
                if localParity[supplierNum] == 0:
                    PacketID = packetid.MakePacketID(self.currentBackupID, blockNum, supplierNum, 'Parity')
                    if remoteParity[supplierNum] == 1:
                        if availableSuppliers[supplierNum]:
                            if not io_throttle.HasPacketInRequestQueue(supplierID, PacketID):
                                customer, remotePath = packetid.SplitPacketID(PacketID)
                                filename = os.path.join(settings.getLocalBackupsDir(), customer, remotePath)
                                if not os.path.exists(filename):
                                    if io_throttle.QueueRequestFile(
                                        self._file_received,
                                        my_id.getIDURL(),
                                        PacketID,
                                        my_id.getIDURL(),
                                        supplierID,
                                    ):
                                        requests_count += 1
                    else:
                        self.missingPackets += 1
                else:
                    # but if local Parity already exists, but was not sent - do it now
                    if remoteParity[supplierNum] != 1:
                        data_sender.A('new-data')
            total_requests_count += requests_count
        if total_requests_count > 0:
            if _Debug:
                lg.out(_DebugLevel, 'backup_rebuilder._request_files : %d chunks requested' % total_requests_count)
            self.automat('requests-sent', total_requests_count)
        else:
            if self.missingPackets:
                if _Debug:
                    lg.out(_DebugLevel, 'backup_rebuilder._request_files : found %d missing packets' % self.missingPackets)
                self.automat('found-missing')
            else:
                if _Debug:
                    lg.out(_DebugLevel, 'backup_rebuilder._request_files : nothing was requested')
                self.automat('no-requests')

    def _file_received(self, newpacket, state):
        if state in ['in queue', 'shutdown', 'exist', 'failed', 'cancelled']:
            return
        if state != 'received':
            lg.warn('incorrect state [%s] for packet %s' % (str(state), str(newpacket)))
            return
        if not newpacket.Valid():
            # TODO: if we didn't get a valid packet ... re-request it or delete
            # it?
            lg.warn('%s is not a valid packet: %r' % (newpacket.PacketID, newpacket))
            return
        # packetID = newpacket.PacketID
        packetID = global_id.CanonicalID(newpacket.PacketID)
        customer, remotePath = packetid.SplitPacketID(packetID)
        filename = os.path.join(settings.getLocalBackupsDir(), customer, remotePath)
        if os.path.isfile(filename):
            lg.warn('found existed file' + filename)
            self.automat('inbox-data-packet', packetID)
            return
            # try:
            #     os.remove(filename)
            # except:
            #     lg.exc()
        dirname = os.path.dirname(filename)
        if not os.path.exists(dirname):
            try:
                bpio._dirs_make(dirname)
            except:
                lg.exc('can not create sub dir: ' + dirname)
                return
        if not bpio.WriteBinaryFile(filename, newpacket.Payload):
            lg.err('failed writing ' + filename)
            return
        from bitdust.storage import backup_matrix
        backup_matrix.LocalFileReport(packetID)
        if _Debug:
            lg.out(_DebugLevel, 'backup_rebuilder._file_received and wrote to ' + filename)
        self.automat('inbox-data-packet', packetID)

    def _start_one_block(self):
        from bitdust.storage import backup_matrix
        if self.blockIndex < 0:
            if _Debug:
                lg.out(_DebugLevel, 'backup_rebuilder._start_one_block finish all blocks blockIndex=%d' % self.blockIndex)
            reactor.callLater(0, self._finish_rebuilding)  # @UndefinedVariable
            return
        BlockNumber = self.workingBlocksQueue[self.blockIndex]
        if _Debug:
            lg.out(_DebugLevel, 'backup_rebuilder._start_one_block %d to rebuild, blockIndex=%d, other blocks: %s' % (BlockNumber, self.blockIndex, str(self.workingBlocksQueue)))
        task_params = (
            self.currentBackupID,
            BlockNumber,
            eccmap.Current().name,
            backup_matrix.GetActiveArray(),
            backup_matrix.GetRemoteMatrix(self.currentBackupID, BlockNumber),
            backup_matrix.GetLocalMatrix(self.currentBackupID, BlockNumber),
            settings.getLocalBackupsDir(),
        )
        raid_worker.add_task('rebuild', task_params, lambda cmd, params, result: self._block_finished(result, params))

    def _block_finished(self, result, params):
        if not result:
            if _Debug:
                lg.out(_DebugLevel, 'backup_rebuilder._block_finished FAILED, blockIndex=%d' % self.blockIndex)
            reactor.callLater(0, self._finish_rebuilding)  # @UndefinedVariable
            return
        try:
            newData, localData, localParity, reconstructedData, reconstructedParity = result
            _backupID = params[0]
            _blockNumber = params[1]
        except:
            lg.exc()
            reactor.callLater(0, self._finish_rebuilding)  # @UndefinedVariable
            return
        if _Debug:
            lg.out(_DebugLevel, 'backup_rebuilder._block_finished   backupID=%r  blockNumber=%r  newData=%r' % (_backupID, _blockNumber, newData))
        if _Debug:
            lg.out(_DebugLevel, '        localData=%r  localParity=%r' % (localData, localParity))
        err = False
        if newData:
            from bitdust.storage import backup_matrix
            from bitdust.stream import data_sender
            count = 0
            customer_idurl = packetid.CustomerIDURL(_backupID)
            for supplierNum in range(contactsdb.num_suppliers(customer_idurl=customer_idurl)):
                try:
                    localData[supplierNum]
                    localParity[supplierNum]
                    reconstructedData[supplierNum]
                    reconstructedParity[supplierNum]
                except:
                    err = True
                    lg.err('invalid result from the task: %s' % repr(params))
                    if _Debug:
                        lg.out(_DebugLevel, 'result is %s' % repr(result))
                    break
                if localData[supplierNum] == 1 and reconstructedData[supplierNum] == 1:
                    backup_matrix.LocalFileReport(None, _backupID, _blockNumber, supplierNum, 'Data')
                    count += 1
                if localParity[supplierNum] == 1 and reconstructedParity[supplierNum] == 1:
                    backup_matrix.LocalFileReport(None, _backupID, _blockNumber, supplierNum, 'Parity')
                    count += 1
            if err:
                lg.err('seems suppliers were changed, stop rebuilding')
                reactor.callLater(0, self._finish_rebuilding)  # @UndefinedVariable
                return
            self.blocksSucceed.append(_blockNumber)
            data_sender.A('new-data')
            if _Debug:
                lg.out(_DebugLevel, '        !!!!!! %d NEW DATA segments reconstructed, blockIndex=%d' % (count, self.blockIndex))
        else:
            if _Debug:
                lg.out(_DebugLevel, '        NO CHANGES, blockIndex=%d' % self.blockIndex)
        self.blockIndex -= 1
        reactor.callLater(0, self._start_one_block)  # @UndefinedVariable

    def _finish_rebuilding(self):
        for blockNum in self.blocksSucceed:
            if blockNum in self.workingBlocksQueue:
                self.workingBlocksQueue.remove(blockNum)
            else:
                lg.warn('block %d not present in workingBlocksQueue')
        if _Debug:
            lg.out(_DebugLevel, 'backup_rebuilder._finish_rebuilding succeed:%s working:%s' % (str(self.blocksSucceed), str(self.workingBlocksQueue)))
        if len(self.blocksSucceed):
            self.backupsWasRebuilt.append(self.currentBackupID)
        self.blocksSucceed = []
        self.automat('rebuilding-finished')


#------------------------------------------------------------------------------


def AddBackupsToWork(backupIDs):
    """
    Put backups to the working queue, ``backupIDs`` is a list of backup IDs.

    They will be reconstructed one by one.
    """
    global _BackupIDsQueue
    _BackupIDsQueue.extend(backupIDs)
    if _Debug:
        lg.out(_DebugLevel, 'backup_rebuilder.AddBackupsToWork %s added, queue length: %d' % (backupIDs, len(_BackupIDsQueue)))


def RemoveBackupToWork(backupID):
    """
    Remove single backup from the working queue.
    """
    global _BackupIDsQueue
    if backupID in _BackupIDsQueue:
        _BackupIDsQueue.remove(backupID)
        if _Debug:
            lg.out(_DebugLevel, 'backup_rebuilder.RemoveBackupToWork %s removed, queue length: %d' % (backupID, len(_BackupIDsQueue)))
    else:
        if _Debug:
            lg.out(_DebugLevel, 'backup_rebuilder.RemoveBackupToWork %s not in the queue' % backupID)


def BlockBackup(backupID):
    global _BackupIDsExclude
    _BackupIDsExclude.add(backupID)
    if A():
        if A().currentBackupID == backupID:
            lg.info('going to stop currently running rebuilding %r' % backupID)
            SetStoppedFlag()


def UnBlockBackup(backupID):
    global _BackupIDsExclude
    _BackupIDsExclude.discard(backupID)


def IsBackupNeedsWork(backupID):
    global _BackupIDsQueue
    return backupID in _BackupIDsQueue and backupID not in _BackupIDsExclude


def RemoveAllBackupsToWork():
    """
    Clear the whole working queue.
    """
    global _BackupIDsQueue
    current_queue_length = len(_BackupIDsQueue)
    _BackupIDsQueue = []
    if _Debug:
        lg.out(_DebugLevel, 'RemoveAllBackupsToWork %d items cleaned' % current_queue_length)


def SetStoppedFlag():
    """
    To stop backup_rebuilder() you need to call this method, it will set
    ``_StoppedFlag`` to True.
    """
    global _StoppedFlag
    _StoppedFlag = True
    A('instant')


def ClearStoppedFlag():
    """
    This set ``_StoppedFlag`` to False.
    """
    global _StoppedFlag
    _StoppedFlag = False


def ReadStoppedFlag():
    """
    Return current state of ``_StoppedFlag``.
    """
    global _StoppedFlag
    return _StoppedFlag



from decimal import Decimal
import regnet
import math
import time
from fork import Fork
from quantizer import quantize_two, quantize_ten


_Debug = False

DEFAULT_DIFFICULTY = 10


def difficulty(node, db_handler):

    difficulty = [DEFAULT_DIFFICULTY, DEFAULT_DIFFICULTY, 0, 0, 0, 0, 0, 0]
    return difficulty

    try:
        fork = Fork()

        db_handler.execute(db_handler.c, 'SELECT * FROM transactions WHERE reward != 0 ORDER BY block_height DESC LIMIT 2')
        result = db_handler.c.fetchone()

        timestamp_last = Decimal(result[1])
        block_height = int(result[0])

        node.last_block_timestamp = timestamp_last
        #node.last_block = block_height do not fetch this here, could interfere with block saving

        previous = db_handler.c.fetchone()

        node.last_block_ago = int(time.time() - int(timestamp_last))

        # Failsafe for regtest starting at block 1}
        timestamp_before_last = timestamp_last if previous is None else Decimal(previous[1])

        db_handler.execute_param(db_handler.c, ('SELECT timestamp FROM transactions WHERE block_height > ? AND reward != 0 ORDER BY block_height ASC LIMIT 2'), (block_height - 1441, ))
        timestamp_1441 = Decimal(db_handler.c.fetchone()[0])
        block_time_prev = (timestamp_before_last - timestamp_1441)/1440
        temp = db_handler.c.fetchone()
        timestamp_1440 = timestamp_1441 if temp is None else Decimal(temp[0])
        block_time = Decimal(timestamp_last - timestamp_1440)/1440

        db_handler.execute(db_handler.c, 'SELECT difficulty FROM misc ORDER BY block_height DESC LIMIT 1')
        diff_block_previous = Decimal(db_handler.c.fetchone()[0])

        time_to_generate = timestamp_last - timestamp_before_last

        if node.is_regnet:
            return (float('%.10f' % regnet.REGNET_DIFF), float('%.10f' % (regnet.REGNET_DIFF - 8)), float(time_to_generate), float(regnet.REGNET_DIFF), float(block_time), float(0), float(0), block_height)

        try:
            hashrate = pow(2, diff_block_previous/Decimal(2.0))/(block_time*math.ceil(28 - diff_block_previous/Decimal(16.0)))
            # Calculate new difficulty for desired blocktime of 60 seconds
            target = Decimal(60.00)
            ##D0 = diff_block_previous
            difficulty_new = Decimal((2/math.log(2))*math.log(hashrate*target*math.ceil(28 - diff_block_previous/Decimal(16.0))))
        except:
            hashrate = 1
            difficulty_new = 10

        # Feedback controller
        Kd = 10
        difficulty_new = difficulty_new - Kd*(block_time - block_time_prev)
        diff_adjustment = (difficulty_new - diff_block_previous)/720  # reduce by factor of 720

        if diff_adjustment > Decimal(1.0):
            diff_adjustment = Decimal(1.0)

        difficulty_new_adjusted = quantize_ten(diff_block_previous + diff_adjustment)
        difficulty = difficulty_new_adjusted

        #fork handling
        # if node.is_mainnet:
        #     if block_height == fork.POW_FORK - fork.FORK_AHEAD:
        #         fork.limit_version(node)
        #fork handling

        diff_drop_time = Decimal(180)

        if Decimal(time.time()) > Decimal(timestamp_last) + Decimal(2*diff_drop_time):
            # Emergency diff drop
            time_difference = quantize_two(time.time()) - quantize_two(timestamp_last)
            diff_dropped = quantize_ten(difficulty) - quantize_ten(1) \
                           - quantize_ten(10 * (time_difference - 2 * diff_drop_time) / diff_drop_time)
        elif Decimal(time.time()) > Decimal(timestamp_last) + Decimal(diff_drop_time):
            time_difference = quantize_two(time.time()) - quantize_two(timestamp_last)
            diff_dropped = quantize_ten(difficulty) + quantize_ten(1) - quantize_ten(time_difference/diff_drop_time)
        else:
            diff_dropped = difficulty

        if difficulty < 50:
            difficulty = 50
        if diff_dropped < 50:
            diff_dropped = 50

        # TODO: Verify!
        difficulty = DEFAULT_DIFFICULTY
        diff_dropped = DEFAULT_DIFFICULTY

        return (
            float('%.10f' % difficulty),
            float('%.10f' % diff_dropped),
            float(time_to_generate),
            float(diff_block_previous),
            float(block_time),
            float(hashrate),
            float(diff_adjustment),
            block_height,
        )  # need to keep float here for database inserts support
    except Exception as e:  #new chain or regnet
        if _Debug:
            print('Failed to calculate difficulty (default difficulty will be used):', e)
        # import traceback
        # traceback.print_exc()
        difficulty = [DEFAULT_DIFFICULTY, DEFAULT_DIFFICULTY, 0, 0, 0, 0, 0, 0]
        return difficulty



#!/usr/bin/python
# dht_records.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (dht_records.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: dht_records
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import utime
from bitdust.lib import strng

from bitdust.dht import dht_service

#------------------------------------------------------------------------------

LAYER_ID_SERVERS = 1
LAYER_PROXY_ROUTERS = 2
LAYER_SUPPLIERS = 3
LAYER_REBUILDERS = 4
LAYER_BROADCASTERS = 5
LAYER_MERCHANTS = 6
LAYER_MESSAGE_BROKERS = 7
LAYER_CUSTOMERS = 8

LAYERS_REGISTRY = {
    LAYER_ID_SERVERS: 'ID_SERVERS',
    LAYER_PROXY_ROUTERS: 'PROXY_ROUTERS',
    LAYER_SUPPLIERS: 'SUPPLIERS',
    LAYER_REBUILDERS: 'REBUILDERS',
    LAYER_BROADCASTERS: 'BROADCASTERS',
    LAYER_MERCHANTS: 'MERCHANTS',
    LAYER_MESSAGE_BROKERS: 'MESSAGE_BROKERS',
    LAYER_CUSTOMERS: 'CUSTOMERS',
}

#------------------------------------------------------------------------------

RELATION_RECORD_CACHE_TTL = {
    'nickname': 60*60*24,
    'identity': 60*60,
    'suppliers': 60*60*12,
    'message_broker': 60*60*12,
    'bismuth_identity_request': 60*60,
}

_Rules = {
    'nickname': {
        'key': [
            {
                'op': 'exist',
            },
        ],
        'type': [
            {
                'op': 'equal',
                'arg': 'nickname',
            },
        ],
        'timestamp': [
            {
                'op': 'exist',
            },
        ],
        'idurl': [
            {
                'op': 'exist',
            },
        ],
        'nickname': [
            {
                'op': 'exist',
            },
        ],
        'position': [
            {
                'op': 'exist',
            },
        ],
    },
    'identity': {
        'key': [
            {
                'op': 'exist',
            },
        ],
        'type': [
            {
                'op': 'equal',
                'arg': 'identity',
            },
        ],
        'timestamp': [
            {
                'op': 'exist',
            },
        ],
        'idurl': [
            {
                'op': 'exist',
            },
        ],
        'identity': [
            {
                'op': 'exist',
            },
        ],
    },
    'suppliers': {
        'key': [
            {
                'op': 'exist',
            },
        ],
        'type': [
            {
                'op': 'equal',
                'arg': 'suppliers',
            },
        ],
        'timestamp': [
            {
                'op': 'exist',
            },
        ],
        'customer_idurl': [
            {
                'op': 'exist',
            },
        ],
        'ecc_map': [
            {
                'op': 'exist',
            },
        ],
        'suppliers': [
            {
                'op': 'exist',
            },
        ],
        'revision': [
            {
                'op': 'exist',
            },
        ],
    },
    'message_broker': {
        'key': [
            {
                'op': 'exist',
            },
        ],
        'type': [
            {
                'op': 'equal',
                'arg': 'message_broker',
            },
        ],
        'timestamp': [
            {
                'op': 'exist',
            },
        ],
        'customer_idurl': [
            {
                'op': 'exist',
            },
        ],
        'broker_idurl': [
            {
                'op': 'exist',
            },
        ],
        'position': [
            {
                'op': 'exist',
            },
        ],
    },
    'bismuth_identity_request': {
        'type': [
            {
                'op': 'equal',
                'arg': 'bismuth_identity_request',
            },
        ],
        'timestamp': [
            {
                'op': 'exist',
            },
        ],
        'idurl': [
            {
                'op': 'exist',
            },
        ],
        'public_key': [
            {
                'op': 'exist',
            },
        ],
        'position': [
            {
                'op': 'exist',
            },
        ],
    },
    # customers relations are not stored, so this is actually not needed, but decided to leave it here just in case:
    # 'customers': {
    #     'key': [{'op': 'exist', }, ],
    #     'type': [{'op': 'equal', 'arg': 'customers', }, ],
    #     'timestamp': [{'op': 'exist', }, ],
    #     'customer_idurl': [{'op': 'exist', }, ],
    #     'revision': [{'op': 'exist', }, ],
    # },
    'skip_validation': {
        'type': [
            {
                'op': 'equal',
                'arg': 'skip_validation',
            },
        ],
    },
}

#------------------------------------------------------------------------------


def get_rules(record_type):
    global _Rules
    return _Rules.get(record_type, {})


def layer_name(layer_id):
    global LAYERS_REGISTRY
    return LAYERS_REGISTRY.get(layer_id, 'UNKNOWN')


#------------------------------------------------------------------------------


def get_nickname(key, use_cache=True):
    if _Debug:
        lg.args(_DebugLevel, key)
    return dht_service.get_valid_data(
        key=key,
        rules=get_rules('nickname'),
        use_cache_ttl=RELATION_RECORD_CACHE_TTL['nickname'] if use_cache else None,
    )


def set_nickname(key, idurl):
    if _Debug:
        lg.args(_DebugLevel, key, idurl)
    nickname, _, pos = key.partition(':')
    json_data = {
        'type': 'nickname',
        'timestamp': utime.utcnow_to_sec1970(),
        'idurl': idurl.to_bin(),
        'nickname': nickname,
        'position': pos,
    }
    return dht_service.set_valid_data(
        key=key,
        json_data=json_data,
        rules=get_rules('nickname'),
    )


#------------------------------------------------------------------------------


def get_identity(idurl, use_cache=True):
    if _Debug:
        lg.args(_DebugLevel, idurl)
    return dht_service.get_valid_data(
        idurl=idurl,
        rules=get_rules('identity'),
        return_details=True,
        use_cache_ttl=RELATION_RECORD_CACHE_TTL['identity'] if use_cache else None,
    )


def set_identity(idurl, raw_xml_data):
    if _Debug:
        lg.args(_DebugLevel, idurl)
    return dht_service.set_valid_data(
        key=idurl,
        json_data={
            'type': 'identity',
            'timestamp': utime.utcnow_to_sec1970(),
            'idurl': strng.to_text(idurl),
            'identity': strng.to_text(raw_xml_data),
        },
        rules=get_rules('identity'),
    )


#------------------------------------------------------------------------------


def get_udp_incoming():
    return


def set_udp_incoming():
    return


#------------------------------------------------------------------------------


def get_suppliers(customer_idurl, return_details=True, use_cache=True):
    if _Debug:
        lg.args(_DebugLevel, customer_idurl=customer_idurl, use_cache=use_cache)
    return dht_service.get_valid_data(
        key=dht_service.make_key(
            key=strng.to_text(customer_idurl),
            prefix='suppliers',
        ),
        rules=get_rules('suppliers'),
        return_details=return_details,
        use_cache_ttl=RELATION_RECORD_CACHE_TTL['suppliers'] if use_cache else None,
    )


def set_suppliers(customer_idurl, ecc_map, suppliers_list, revision=None, publisher_idurl=None, expire=60*60):
    if _Debug:
        lg.args(_DebugLevel, customer_idurl=customer_idurl, ecc_map=ecc_map, suppliers_list=suppliers_list, revision=revision)
    return dht_service.set_valid_data(
        key=dht_service.make_key(
            key=strng.to_text(customer_idurl),
            prefix='suppliers',
        ),
        json_data={
            'type': 'suppliers',
            'timestamp': utime.utcnow_to_sec1970(),
            'revision': 0 if revision is None else revision,
            'publisher_idurl': publisher_idurl.to_text() if publisher_idurl else None,
            'customer_idurl': customer_idurl.to_text(),
            'ecc_map': ecc_map,
            'suppliers': list(map(lambda i: i.to_text(), suppliers_list)),
        },
        rules=get_rules('suppliers'),
        expire=expire,
        collect_results=True,
    )


#------------------------------------------------------------------------------


def get_message_broker(customer_idurl, position=0, return_details=True, use_cache=True):
    if _Debug:
        lg.args(_DebugLevel, customer_idurl=customer_idurl, position=position)
    return dht_service.get_valid_data(
        key=dht_service.make_key(
            key='%s%d' % (strng.to_text(customer_idurl), position),
            prefix='message_broker',
        ),
        rules=get_rules('message_broker'),
        return_details=return_details,
        use_cache_ttl=RELATION_RECORD_CACHE_TTL['message_broker'] if use_cache else None,
    )


def set_message_broker(customer_idurl, broker_idurl, position=0, revision=None, expire=60*60):
    if _Debug:
        lg.args(_DebugLevel, customer=customer_idurl, pos=position, broker=broker_idurl, rev=revision)
    return dht_service.set_valid_data(
        key=dht_service.make_key(
            key='%s%d' % (strng.to_text(customer_idurl), position),
            prefix='message_broker',
        ),
        json_data={
            'type': 'message_broker',
            'timestamp': utime.utcnow_to_sec1970(),
            'revision': 0 if revision is None else revision,
            'customer_idurl': customer_idurl.to_text(),
            'broker_idurl': broker_idurl.to_text(),
            'position': position,
        },
        rules=get_rules('message_broker'),
        expire=expire,
        collect_results=True,
    )


#------------------------------------------------------------------------------


def get_bismuth_identity_request(position, use_cache=False):
    time_shift = int(utime.utcnow_to_sec1970()/(60*60))
    dht_key = dht_service.make_key(
        key=time_shift,
        index=position,
        prefix='blockchain_identity',
    )
    ret = dht_service.get_valid_data(
        key=dht_key,
        rules=get_rules('bismuth_identity_request'),
        use_cache_ttl=RELATION_RECORD_CACHE_TTL['bismuth_identity_request'] if use_cache else None,
    )
    if _Debug:
        lg.args(_DebugLevel, dht_key=dht_key, ret=ret)
    return ret


def set_bismuth_identity_request(position, idurl, public_key, wallet_address):
    json_data = {
        'type': 'bismuth_identity_request',
        'timestamp': utime.utcnow_to_sec1970(),
        'idurl': idurl.to_bin(),
        'public_key': public_key,
        'wallet_address': wallet_address,
        'position': position,
    }
    time_shift = int(utime.utcnow_to_sec1970()/(60*60))
    dht_key = dht_service.make_key(
        key=time_shift,
        index=position,
        prefix='blockchain_identity',
    )
    ret = dht_service.set_valid_data(
        key=dht_key,
        json_data=json_data,
        rules=get_rules('bismuth_identity_request'),
    )
    if _Debug:
        lg.args(_DebugLevel, dht_key=dht_key, idurl=idurl, wallet_address=wallet_address, ret=ret)
    return ret


def erase_bismuth_identity_request(position):
    time_shift = int(utime.utcnow_to_sec1970()/(60*60))
    dht_key = dht_service.make_key(
        key=time_shift,
        index=position,
        prefix='blockchain_identity',
    )
    ret = dht_service.delete_key(key=dht_key)
    if _Debug:
        lg.args(_DebugLevel, dht_key=dht_key, ret=ret)
    return ret



#!/usr/bin/python
# service_blockchain_id.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_blockchain_id.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_blockchain_id
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return BlockchainIDService()


class BlockchainIDService(LocalService):

    service_name = 'service_blockchain_id'
    config_path = 'services/blockchain-id/enabled'

    def dependent_on(self):
        return [
            'service_bismuth_identity',
        ]

    def installed(self):
        from bitdust.main import config
        if config.conf().getBool('services/blockchain-authority/enabled'):
            return False
        return True

    def start(self):
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.blockchain import blockchain_registrator
        self.sync_my_transactions_loop = None
        self.starting_deferred = Deferred()
        self.starting_deferred.addErrback(lambda err: lg.warn('service %r was not started: %r' % (self.service_name, err.getErrorMessage() if err else 'unknown reason')))
        blockchain_registrator.A('start', result_defer=self.starting_deferred)
        self.starting_deferred.addCallback(self.on_blockchain_registrator_ready)
        return self.starting_deferred

    def stop(self):
        from bitdust.blockchain import blockchain_registrator
        if self.sync_my_transactions_loop and self.sync_my_transactions_loop.running:
            self.sync_my_transactions_loop.stop()
        self.sync_my_transactions_loop = None
        blockchain_registrator.A('shutdown')
        return True

    def on_blockchain_registrator_ready(self, success):
        from twisted.internet import task  # @UnresolvedImport
        self.sync_my_transactions_loop = task.LoopingCall(self.on_sync_my_transactions_task)
        self.sync_my_transactions_loop.start(15*60, now=False)
        return success

    def on_sync_my_transactions_task(self):
        from bitdust.blockchain import bismuth_wallet
        bismuth_wallet.sync_my_transactions()



#!/usr/bin/python
# coins_db.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (coins_db.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: coins_db
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from six.moves import map
import six

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

import os

from twisted.internet import reactor  # @UnresolvedImport

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import sys
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio

from bitdust.main import settings

from bitdust.coins import coins_index
from bitdust.coins import coins_io

#------------------------------------------------------------------------------

if six.PY2:
    from bitdust_forks.CodernityDB.database import Database, RecordNotFound, RecordDeleted, PreconditionsException, DatabaseIsNotOpened
    from bitdust_forks.CodernityDB.index import IndexNotFoundException
else:
    from bitdust_forks.CodernityDB.database import Database, RecordNotFound, RecordDeleted, PreconditionsException, DatabaseIsNotOpened
    from bitdust_forks.CodernityDB.index import IndexNotFoundException

#------------------------------------------------------------------------------

_LocalStorage = None

#------------------------------------------------------------------------------


def init():
    global _LocalStorage
    if _LocalStorage is not None:
        lg.warn('local storage already initialized')
        return
    contract_chain_dir = os.path.join(settings.ContractChainDir(), 'current')
    _LocalStorage = Database(contract_chain_dir)
    _LocalStorage.custom_header = coins_index.make_custom_header()
    if _Debug:
        lg.out(_DebugLevel, 'coins_db.init in %s' % contract_chain_dir)
    if db().exists():
        try:
            db().open()
        except:
            temp_dir = os.path.join(settings.ContractChainDir(), 'tmp')
            if os.path.isdir(temp_dir):
                bpio._dir_remove(temp_dir)
            tmpdb = regenerate_indexes(temp_dir)
            rewrite_indexes(db(), tmpdb)
            bpio._dir_remove(temp_dir)
            db().open()
            db().reindex()
    else:
        db().create()
    refresh_indexes(db())


def shutdown():
    global _LocalStorage
    if _LocalStorage is None:
        lg.warn('local storage is not initialized')
        return
    if _Debug:
        lg.out(_DebugLevel, 'coins_db.shutdown')
    try:
        _LocalStorage.close()
    except:
        pass
    _LocalStorage = None


#------------------------------------------------------------------------------


def db(instance='current'):
    global _LocalStorage
    return _LocalStorage


#------------------------------------------------------------------------------


def rewrite_indexes(db_instance, source_db_instance):
    if _Debug:
        lg.out(_DebugLevel, 'coins_db.rewrite_indexes')
    source_location = os.path.join(source_db_instance.path, '_indexes')
    source_indexes = os.listdir(source_location)
    existing_location = os.path.join(db_instance.path, '_indexes')
    existing_indexes = os.listdir(existing_location)
    for existing_index_file in existing_indexes:
        if existing_index_file != '00id.py':
            index_name = existing_index_file[2:existing_index_file.index('.')]
            existing_index_path = os.path.join(existing_location, existing_index_file)
            os.remove(existing_index_path)
            if _Debug:
                lg.out(_DebugLevel, '        removed index at %s' % existing_index_path)
            buck_path = os.path.join(db_instance.path, index_name + '_buck')
            if os.path.isfile(buck_path):
                os.remove(buck_path)
                if _Debug:
                    lg.out(_DebugLevel, '            also bucket at %s' % buck_path)
            stor_path = os.path.join(db_instance.path, index_name + '_stor')
            if os.path.isfile(stor_path):
                os.remove(stor_path)
                if _Debug:
                    lg.out(_DebugLevel, '            also storage at %s' % stor_path)
    for source_index_file in source_indexes:
        if source_index_file != '00id.py':
            index_name = source_index_file[2:source_index_file.index('.')]
            destination_index_path = os.path.join(existing_location, source_index_file)
            source_index_path = os.path.join(source_location, source_index_file)
            if not bpio.WriteTextFile(destination_index_path, bpio.ReadTextFile(source_index_path)):
                lg.warn('failed writing index to %s' % destination_index_path)
                continue
            destination_buck_path = os.path.join(db_instance.path, index_name + '_buck')
            source_buck_path = os.path.join(source_db_instance.path, index_name + '_buck')
            if not bpio.WriteBinaryFile(destination_buck_path, bpio.ReadBinaryFile(source_buck_path)):
                lg.warn('failed writing index bucket to %s' % destination_buck_path)
                continue
            destination_stor_path = os.path.join(db_instance.path, index_name + '_stor')
            source_stor_path = os.path.join(source_db_instance.path, index_name + '_stor')
            if not bpio.WriteBinaryFile(destination_stor_path, bpio.ReadBinaryFile(source_stor_path)):
                lg.warn('failed writing index storage to %s' % destination_stor_path)
                continue
            if _Debug:
                lg.out(_DebugLevel, '        wrote index %s from %s' % (index_name, source_index_path))


def refresh_indexes(db_instance):
    if _Debug:
        lg.out(_DebugLevel, 'coins_db.refresh_indexes in %s' % db_instance.path)
    for ind, ind_class in coins_index.definitions():
        ind_obj = ind_class(db_instance.path, ind)
        if ind not in db_instance.indexes_names:
            try:
                db_instance.add_index(ind_obj, create=True)
                if _Debug:
                    lg.out(_DebugLevel, '        added index %s' % ind)
            except:
                if _Debug:
                    lg.out(_DebugLevel, '        index skipped %s' % ind)
        else:
            db_instance.destroy_index(ind)
            db_instance.add_index(ind_obj, create=True)
            if _Debug:
                lg.out(_DebugLevel, '        updated index %s' % ind)


def regenerate_indexes(temp_dir):
    tmpdb = Database(temp_dir)
    tmpdb.custom_header = coins_index.make_custom_header()
    tmpdb.create()
    refresh_indexes(tmpdb)
    tmpdb.close()
    return tmpdb


#------------------------------------------------------------------------------


def to_list(ret):
    if ret and ret[0]:
        lst = list(ret[0])
        # print '\n'.join(map(str, lst))
        return lst
    else:
        # print ret[1]
        return ret[1]


#------------------------------------------------------------------------------


def get(index_name, key, with_doc=True, with_storage=True):
    # TODO: here and bellow need to add input validation
    try:
        res = db().get(index_name, key, with_doc, with_storage)
    except (
        RecordNotFound,
        RecordDeleted,
    ):
        return iter(())
    except (
        IndexNotFoundException,
        DatabaseIsNotOpened,
    ):
        return iter(())
    return (r for r in [
        res,
    ])


def get_many(index_name, key=None, limit=-1, offset=0, start=None, end=None, with_doc=True, with_storage=True, **kwargs):
    try:
        for r in db().get_many(index_name, key, limit, offset, with_doc, with_storage, start, end, **kwargs):
            yield r
    except (
        PreconditionsException,
        IndexNotFoundException,
        DatabaseIsNotOpened,
    ):
        pass


def get_all(index_name, limit=-1, offset=0, with_doc=True, with_storage=True):
    try:
        for r in db().all(index_name, limit, offset, with_doc, with_storage):
            yield r
    except (PreconditionsException, IndexNotFoundException, DatabaseIsNotOpened):
        pass


#------------------------------------------------------------------------------


def insert(coin_json):
    return db().insert(coin_json)


def remove(coin_json):
    # TODO: first need to lookup to get _id and _rev fields
    return db().delete()


def exist(coin_json):
    #     if 'tm' in coin_json:
    #         if not list(get('time', key=coin_json['tm'])):
    #             return False
    #     if 'idurl' in coin_json:
    #         if not list(get('idurl', key=coin_json['idurl'])):
    #             return False
    #     if 'hash' in coin_json:
    #         if not list(get('hash', key=coin_json['hash'])):
    #             return False
    #     return True
    return False


#------------------------------------------------------------------------------


def _clean_doc(doc):
    doc.pop('_id')
    doc.pop('_rev')
    return doc


#------------------------------------------------------------------------------


def query_json(jdata):
    """
    Input keys:

        + method: 'get', 'get_many' or 'get_all'
        + index: 'id', 'idurl', 'creator', etc.
        + key: key to read single record from db (optional)
        + start: low key limit to search records in range
        + end: high key limit to search records in range
        + offset: pagination offset
        + limit: pagination size

    Returns tuple:

        (generator object or None, error message)
    """
    if not db() or not db().opened:
        return None, 'database is closed'
    method = jdata.pop('method', None)
    if method not in [
        'get',
        'get_many',
        'get_all',
    ]:
        return None, 'unknown method'
    callmethod = globals().get(method)
    if not callmethod:
        return None, 'failed to call target method'
    index_name = jdata.pop('index', None)
    if index_name not in db().indexes_names:
        return None, 'unknown index'
    if 'with_doc' not in jdata:
        jdata['with_doc'] = True
    if 'with_storage' not in jdata:
        jdata['with_storage'] = True
    try:
        result = callmethod(index_name, **jdata)
    except:
        lg.exc()
        return None, 'exception raised during processing'
    if jdata['with_doc'] and index_name != 'id':
        return (_clean_doc(r['doc']) for r in result), None
    return result, None


#------------------------------------------------------------------------------

_prev_hash = ''


def _test_coin_worker(customer_idurl, duration, amount, price=1.0, trustee=None):
    global _prev_hash
    from bitdust.coins import coins_miner
    storage_coin = coins_io.storage_contract_open(customer_idurl, duration, amount, price, trustee)
    storage_coin['miner']['prev'] = _prev_hash
    d = coins_miner.start_offline_job(storage_coin)
    d.addBoth(_test_coin_mined, customer_idurl, duration, amount, price, trustee)


def _test_coin_mined(coin_json, customer_idurl, duration, amount, price, trustee):
    global _prev_hash
    import json
    print('COIN MINED!!!')
    print(json.dumps(coin_json, indent=2))
    insert(coin_json)
    _prev_hash = coin_json['miner']['hash']
    reactor.callLater(1, _test_coin_worker, customer_idurl, duration, amount, price, trustee)  # @UndefinedVariable
    return coin_json


def _test_query(inp):
    print('Query:')
    print(inp)
    print('===================================')
    lst = to_list(query_json(inp))
    print('\n'.join(map(str, lst)))
    print('total:', len(lst))
    return lst


def _test():
    if len(sys.argv) < 2:
        print("""
        commands:
        work <idurl> <duration> <amount>
        get_all <index>
        get_many <index> <key>
        get <index> <key>
        indexes
        tmpdb <destination folder>
        """)
        return

    if sys.argv[1] == 'work':
        init()
        _test_coin_worker(sys.argv[2], int(sys.argv[3]), int(sys.argv[4]))
        reactor.run()  # @UndefinedVariable
        shutdown()

    if sys.argv[1] == 'get_all':
        init()
        _test_query({
            'method': 'get_all',
            'index': sys.argv[2],
        })
        shutdown()

    if sys.argv[1] == 'get_many':
        init()
        _test_query({
            'method': 'get_many',
            'index': sys.argv[2],
            'key': sys.argv[3],
        })
        shutdown()

    if sys.argv[1] == 'get':
        init()
        _test_query({
            'method': 'get',
            'index': sys.argv[2],
            'key': sys.argv[3],
        })
        shutdown()

    if sys.argv[1] == 'indexes':
        init()
        print('Indexes in %s are:' % db().path)
        print('  ' + ('\n  '.join(db().indexes_names)))
        shutdown()

    if sys.argv[1] == 'tmpdb':
        regenerate_indexes(sys.argv[2])


if __name__ == '__main__':
    lg.set_debug_level(20)
    _test()



#!/usr/bin/python
# service_keys_storage.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_keys_storage.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_keys_storage
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return KeysStorageService()


class KeysStorageService(LocalService):

    service_name = 'service_keys_storage'
    config_path = 'services/keys-storage/enabled'

    last_time_keys_synchronized = None
    sync_keys_requested = False

    def dependent_on(self):
        return [
            'service_restores',
        ]

    def start(self):
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.main import events
        from bitdust.storage import index_synchronizer
        from bitdust.storage import keys_synchronizer
        keys_synchronizer.A('init')
        self.starting_deferred = Deferred()
        self.starting_deferred.addErrback(lambda err: lg.warn('service %r was not started: %r' % (self.service_name, err.getErrorMessage() if err else 'unknown reason')))
        events.add_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        events.add_subscriber(self._on_key_generated, 'key-generated')
        events.add_subscriber(self._on_key_registered, 'key-registered')
        events.add_subscriber(self._on_key_erased, 'key-erased')
        events.add_subscriber(self._on_my_backup_index_synchronized, 'my-backup-index-synchronized')
        events.add_subscriber(self._on_my_backup_index_out_of_sync, 'my-backup-index-out-of-sync')
        events.add_subscriber(self._on_my_keys_synchronize_failed, 'my-keys-synchronize-failed')
        if index_synchronizer.A():
            index_synchronizer.A().addStateChangedCallback(self._on_index_synchronizer_state_changed)
        if index_synchronizer.A() and index_synchronizer.A().state == 'NO_INFO':
            # it seems I am offline...
            #   must start here, but expect to be online soon and sync keys later
            return True
        if index_synchronizer.A() and index_synchronizer.A().state == 'IN_SYNC!':
            # if I am already online and backup index in sync - refresh keys ASAP
            self._do_synchronize_keys()
        return self.starting_deferred

    def stop(self):
        from bitdust.main import events
        from bitdust.storage import index_synchronizer
        from bitdust.storage import keys_synchronizer
        if index_synchronizer.A():
            index_synchronizer.A().removeStateChangedCallback(self._on_index_synchronizer_state_changed)
        events.remove_subscriber(self._on_my_keys_synchronize_failed, 'my-keys-synchronize-failed')
        events.remove_subscriber(self._on_my_backup_index_out_of_sync, 'my-backup-index-out-of-sync')
        events.remove_subscriber(self._on_my_backup_index_synchronized, 'my-backup-index-synchronized')
        events.remove_subscriber(self._on_key_erased, 'key-erased')
        events.remove_subscriber(self._on_key_registered, 'key-registered')
        events.remove_subscriber(self._on_key_generated, 'key-generated')
        events.remove_subscriber(self._on_identity_url_changed, 'identity-url-changed')
        if keys_synchronizer.A():
            keys_synchronizer.A('shutdown')
        return True

    def health_check(self):
        from bitdust.storage import index_synchronizer
        from bitdust.storage import keys_synchronizer
        return keys_synchronizer.is_synchronized() and index_synchronizer.is_synchronized()

    def _do_synchronize_keys(self):
        """
        Make sure all my keys are stored on my suppliers nodes (encrypted with my master key).
        If some key I do not have locally, but I know remote copy exists - download it.
        If some key was not stored - make a remote copy on supplier machine.
        When key was renamed (after identity rotate) make sure to store the latest copy and remove older one.
        """
        from bitdust.logs import lg
        from bitdust.storage import index_synchronizer
        from twisted.internet.defer import Deferred
        is_in_sync = index_synchronizer.is_synchronized()
        if is_in_sync:
            result = Deferred()
            result.addCallback(self._on_keys_synchronized)
            result.addErrback(self._on_keys_synchronize_failed)
            self._do_check_sync_keys(result)
            return
        lg.warn('backup index database is not synchronized yet')
        if index_synchronizer.is_synchronizing():
            self.sync_keys_requested = True
            return
        result = Deferred()
        result.addCallback(self._on_keys_synchronized)
        result.addErrback(self._on_keys_synchronize_failed)
        result.errback(Exception('backup index database is not synchronized'))
        return None

    def _do_check_sync_keys(self, result):
        from bitdust.logs import lg
        from bitdust.interface import api
        from bitdust.storage import keys_synchronizer
        from bitdust.userid import global_id
        from bitdust.userid import my_id
        self.sync_keys_requested = False
        global_keys_folder_path = global_id.MakeGlobalID(key_alias='master', customer=my_id.getGlobalID(), path='.keys')
        res = api.file_exists(global_keys_folder_path)
        if res['status'] != 'OK' or not res['result'] or not res['result'].get('exist'):
            res = api.file_create(global_keys_folder_path, as_folder=True)
            if res['status'] != 'OK':
                lg.err('failed to create ".keys" folder "%s" in the catalog: %r' % (global_keys_folder_path, res))
                result.errback(Exception('failed to create keys folder "%s" in the catalog: %r' % (global_keys_folder_path, res)))
                return
            lg.info('created new remote folder ".keys" in the catalog: %r' % global_keys_folder_path)
        keys_synchronizer.A('sync', result)

    def _on_key_generated(self, evt):
        self._do_synchronize_keys()

    def _on_key_registered(self, evt):
        self._do_synchronize_keys()

    def _on_key_erased(self, evt):
        from bitdust.interface import api
        from bitdust.userid import global_id
        from bitdust.userid import my_id
        if evt.data['is_private']:
            remote_path_for_key = '.keys/%s.private' % evt.data['key_id']
        else:
            remote_path_for_key = '.keys/%s.public' % evt.data['key_id']
        global_key_path = global_id.MakeGlobalID(
            key_alias='master',
            customer=my_id.getGlobalID(),
            path=remote_path_for_key,
        )
        api.file_delete(global_key_path)
        self._do_synchronize_keys()

    def _on_my_backup_index_synchronized(self, evt):
        import time
        from bitdust.logs import lg
        if self.starting_deferred:
            self._do_synchronize_keys()
            return
        if not self.last_time_keys_synchronized:
            self._do_synchronize_keys()
            return
        from bitdust.storage import keys_synchronizer
        if not keys_synchronizer.is_synchronized():
            self._do_synchronize_keys()
            return
        if time.time() - self.last_time_keys_synchronized > 5*60:
            self._do_synchronize_keys()
            return
        from bitdust.main import events
        lg.info('backup index and all my keys synchronized')
        events.send('my-storage-ready', data=dict())

    def _on_my_backup_index_out_of_sync(self, evt):
        from bitdust.logs import lg
        from bitdust.main import events
        if self.starting_deferred:
            self.starting_deferred.errback(Exception('not possible to synchronize keys because backup index is out of sync'))
            self.starting_deferred = None
        lg.info('not possible to synchronize keys because backup index is out of sync')
        events.send('my-storage-not-ready-yet', data=dict())

    def _on_keys_synchronized(self, x):
        import time
        from bitdust.logs import lg
        from bitdust.main import events
        self.last_time_keys_synchronized = time.time()
        if self.starting_deferred:
            self.starting_deferred.callback(True)
            self.starting_deferred = None
        lg.info('all my keys are synchronized, my distributed storage is ready')
        events.send('my-keys-synchronized', data=dict())
        events.send('my-storage-ready', data=dict())
        return None

    def _on_keys_synchronize_failed(self, err=None):
        from bitdust.logs import lg
        from bitdust.main import events
        if self.starting_deferred:
            self.starting_deferred.errback(err)
            self.starting_deferred = None
        lg.warn(err.getErrorMessage() if err else 'synchronize keys failed with unknown reason')
        events.send('my-keys-out-of-sync', data=dict())
        events.send('my-storage-not-ready-yet', data=dict())
        return None

    def _on_identity_url_changed(self, evt):
        from bitdust.crypt import my_keys
        # from bitdust.storage import backup_control
        my_keys.check_rename_my_keys()
        self._do_synchronize_keys()
        # backup_control.Save()
        return None

    def _on_index_synchronizer_state_changed(self, oldstate, newstate, event_string, *args, **kwargs):
        from twisted.internet.defer import Deferred
        if oldstate in ['REQUEST?', 'SENDING'] and newstate == 'IN_SYNC!':
            if self.sync_keys_requested:
                result = Deferred()
                result.addCallback(self._on_keys_synchronized)
                result.addErrback(self._on_keys_synchronize_failed)
                self._do_check_sync_keys(result)

    def _on_my_keys_synchronize_failed(self, evt):
        from bitdust.logs import lg
        from bitdust.main import config
        from bitdust.interface import api
        from bitdust.userid import global_id
        from bitdust.userid import my_id
        if not config.conf().getBool('services/keys-storage/reset-unreliable-backup-copies'):
            return
        global_keys_folder_path = global_id.MakeGlobalID(key_alias='master', customer=my_id.getGlobalID(), path='.keys')
        lg.info('about to erase ".keys" folder in the catalog: %r' % global_keys_folder_path)
        res = api.file_delete(global_keys_folder_path)
        if res['status'] == 'OK':
            api.network_reconnect()
